/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include "libacq.h"


void init_acquisition(struct acq_workspace *ws) {

  //init the reference clock
  ws->refclock=malloc(4096*sizeof(char));
  strcpy(ws->refclock,"noclock");
}


void deinit_acquisition(struct acq_workspace *ws) {
  struct acqunit_workspace *auws;
  struct data_descriptor *current_ctrl;

  auws=ws->acqunits;
  while(auws!=NULL) {
    del_acq_unit(auws);
    auws=auws->next;
  }

  ws->acqunits=NULL;
  free(ws->refclock);

  //empty the control buffers
  printf("empty the control buffers\n");
  current_ctrl=ws->begin_ctrl;
  while(current_ctrl!=ws->end_ctrl) {
    if (current_ctrl->id!=-1) {
      leave_generic_buffer(ws->bpool,current_ctrl->generic_buffer);
      current_ctrl->id=-1;
    }
    current_ctrl=current_ctrl->next;
  }
  ws->begin_ctrl=ws->end_ctrl;

  printf("deinit_acquisition done\n");
}


void print_packet(struct generic_buffer *acq_buffer) {
  int i;
  for(i=0;i<acq_buffer->size;i++)
    printf("%x ",acq_buffer->data[i]);
  printf("\n");
}



int insert_new_data_descriptor(struct acq_workspace *ws,struct generic_buffer *acq_buffer,struct acqunit_workspace *auws,int acqid) {
  pthread_mutex_lock(ws->data_descriptor_mutex);
  if (ws->current_data->id<0) {
    //filling a data descriptor
    ws->current_data->generic_buffer=acq_buffer;
    //assign the acquisition unit workspace to this packet
    ws->current_data->auws=auws;
    ws->current_data->acqid=acqid;
    //set the id to the first consumer
    ws->current_data->id=MAXID;
    //going to the next descriptor
    ws->current_data=ws->current_data->next;
    //set statistics
    set_stats(ws->stats,"nb_data_pkts",SOP_PLUS,1);
    pthread_mutex_unlock(ws->data_descriptor_mutex);
    return 1;
  } else {
    printf("no data descriptor available\n");
    leave_generic_buffer(ws->bpool,acq_buffer);
    set_stats(ws->stats,"nb_lost_pkts",SOP_PLUS,1);
    pthread_mutex_unlock(ws->data_descriptor_mutex);
    return 0;
  }
}


int insert_new_ctrl_descriptor(struct acq_workspace *ws,struct generic_buffer *acq_buffer,struct acqunit_workspace *auws) {
  pthread_mutex_lock(ws->ctrl_descriptor_mutex);
  ws->end_ctrl->generic_buffer=acq_buffer;
  ws->end_ctrl->id=1;
  ws->end_ctrl->auws=auws;
  ws->end_ctrl=ws->end_ctrl->next;
  if (ws->end_ctrl==ws->begin_ctrl) {
    //printf("dropping an old packet\n");
    leave_generic_buffer(ws->bpool,ws->begin_ctrl->generic_buffer);
    ws->begin_ctrl=ws->begin_ctrl->next;
  } 
  set_stats(ws->stats,"nb_ctrl_pkts",SOP_PLUS,1); 
  pthread_mutex_unlock(ws->ctrl_descriptor_mutex);
  return 1;
}


void *acquisition(void *args) {
  struct acqunit_workspace *auws=(struct acqunit_workspace *)args;
  struct acq_workspace *ws=auws->globalws;
  int active;
  int localactive=0;
  int size;
  int res,wres;
  int acqid=0;
  unsigned char packet[1024];
  struct generic_buffer *acq_buffer;
  struct data_descriptor *ddcur;
  
  //infinite loop
  while(1) {

    //if deinit is engaged, stop the thread
    if (auws->deinit==1) {
      printf("acquisition deinit order detected\n");
      break;
    }

    //get the flag to see if the acquisition should be active or not
    pthread_mutex_lock(auws->acq_active_mutex);
    active=auws->acq_active;
    pthread_mutex_unlock(auws->acq_active_mutex);
    //compare with reality
    if (active!=localactive) {
      debug("active=%d localactive=%d\n",active,localactive);
      //if necessary do the action corresponding to the flag
      if (!active) {
        (auws->stop_acq)(auws->specws);
        localactive=0;
        usleep(2000);
        continue;
      } else {
        (auws->start_acq)(auws->specws);
        localactive=1;
      }
    }
    //if not active just wait
    if (!active) {
      usleep(2000);
      continue;
    }

    //acquisition part
    //find a free buffer
    acq_buffer=get_generic_buffer(ws->bpool);
    if (acq_buffer!=NULL) {
      //printf("acquire one block\n");
      size=(auws->acquire_one_block_acq)(auws->specws,acq_buffer->data,MAX_PACKET_SIZE);
      //printf("block size=%d\n",size);
      if (size<0) {
        printf("acquire_one_block returns %d\n",size);
        printf("read error : stoping acquisition\n");
        stop_acquisition(auws);
        leave_generic_buffer(ws->bpool,acq_buffer);
        continue;
      }
      if (size==0) {
        leave_generic_buffer(ws->bpool,acq_buffer);
        continue;
      }
      acq_buffer->size=size;
      //the acquisition is done : call prefilter to know if it is junk,data or control
      res=(auws->prefilter_uncap)(auws->uncapws,acq_buffer->data,acq_buffer->size);
      switch (res) {
      case PREF_JUNK: 
        //drop junk packet
        //printf("treating junk packet\n");
        leave_generic_buffer(ws->bpool,acq_buffer);
        break;
      case PREF_DATA: 
        //warn the scheduler that we go into a bunch of acquisition
        ws->packet_detector=1;
        //give the data packet to the consumers chain
        //printf("treating data packet\n");
        insert_new_data_descriptor(ws,acq_buffer,auws,acqid);
        acqid++;
        //awake the first consumer
        wres=write(ws->consumers[MAXID]->cons_pipe[1],"1",1);
        if (wres==0)
          printf("nothing written\n");
        if (wres==-1) {
          perror("write");
        }
        break;
      case PREF_CTRL: 
        //store the control packet in a circular buffer pool
        //printf("treating ctrl packet\n");
        insert_new_ctrl_descriptor(ws,acq_buffer,auws);
        break;
      default: //should never happen
        printf("problem with prefilter : return value = %d\n",res);
        leave_generic_buffer(ws->bpool,acq_buffer);
      }
    } else { //no room left
      printf("bufferpool is full\n");
      size=(auws->acquire_one_block_acq)(auws->specws,packet,4096);
      if (size<0) {
        printf("acquire_one_block returns %d\n",size);
        printf("read error : stoping acquisition\n");
        stop_acquisition(auws);
      }
      set_stats(ws->stats,"nb_lost_pkts",SOP_PLUS,1);
    } 
  }


  //deinit code

  //deinit uncap and acq modules
  printf("calling acq deinit\n");
  auws->deinit_acq(auws->specws); 
  printf("calling uncap deinit\n");
  auws->deinit_uncap(auws->uncapws);

  //TODO add a remove of bad packets
  //force all buffer to be free
  ddcur=ws->current_data->prev;
  while(ddcur!=ws->current_data) {
   if (ddcur->id!=-1) {
     if (ddcur->auws==auws) {
       ddcur->auws=NULL;
     }
   }
   ddcur=ddcur->prev;
  }
  
  printf("exiting the acquisition thread\n");
  auws->deinitialized=1;
  return NULL;
}
