/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include "libacq.h"

void *load_func(void *lib,const char * name) {
  void (*func)();
  func=dlsym(lib,name);
  return func;
}


struct acqunit_workspace * make_acq_unit(char *name,int nb_streams,int fid,char * acq_lib,char * uncap_lib,char *datadir,char *convert_script,int timeout,int replay,struct acq_workspace *ws) {
  struct acqunit_workspace * result=malloc(sizeof(struct acqunit_workspace));
  memset(result,0,sizeof(struct acqunit_workspace));
  int i;

  result->nb_streams=nb_streams;
  result->deinit=0;
  result->fid=fid;
  result->deinitialized=0;
  result->name=strdup(name);
  result->timeout=timeout;
  result->replay_enabled=replay;
  result->replay_file=-1;
  printf("name is %s\n",name);
  
  //fill the pointer with the acquisition library functions
  result->lib_acq=dlopen(acq_lib,RTLD_NOW|RTLD_LOCAL|RTLD_DEEPBIND);
  if (!result->lib_acq) {
    printf("dlopen error : %s\n",dlerror()); 
    free(result);
    return NULL;
  }
  //fill the pointer with the uncap library functions
  result->lib_uncap=dlopen(uncap_lib,RTLD_NOW|RTLD_LOCAL|RTLD_DEEPBIND);
  if (!result->lib_uncap) {
    printf("dlopen error : %s\n",dlerror()); 
    dlclose(result->lib_acq);
    free(result);
    return NULL;
  }
  result->init_acq=load_func(result->lib_acq,"init_acq");
  result->deinit_acq=load_func(result->lib_acq,"deinit_acq");
  result->start_acq=load_func(result->lib_acq,"start_acq");
  result->stop_acq=load_func(result->lib_acq,"stop_acq");
  result->send_acq=load_func(result->lib_acq,"send_acq");
  result->acquire_one_block_acq=load_func(result->lib_acq,"acquire_one_block");
  if (result->init_acq==NULL || result->deinit_acq==NULL || result->start_acq==NULL || result->stop_acq==NULL || result->acquire_one_block_acq==NULL) {
    printf("basic function for acquisition are not loaded...giving up\n");
    submod_exit(1);
  }
 
  result->init_uncap=load_func(result->lib_uncap,"init");
  result->deinit_uncap=load_func(result->lib_uncap,"deinit");
  result->prefilter_uncap=load_func(result->lib_uncap,"prefilter");
  result->uncap_uncap=load_func(result->lib_uncap,"uncap");
  result->select_packet_uncap=load_func(result->lib_uncap,"select_packet");
  if (result->init_uncap==NULL || result->deinit_uncap==NULL || result->prefilter_uncap==NULL || result->uncap_uncap==NULL || result->select_packet_uncap==NULL) {
    printf("basic function for uncap are not loaded...giving up\n");
    submod_exit(1);
  }
  
  result->globalws=ws;
  result->datadir=strdup(datadir);
  result->convert_script=strdup(convert_script);
  result->file_mutex=malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(result->file_mutex,NULL);

  //init the active variable and its mutex
  result->acq_active_mutex=malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(result->acq_active_mutex,NULL);
  result->acq_active=0;

   //initializing the streams
  debug("initializing the streams\n");
  result->streams=malloc(nb_streams*sizeof(struct stream_workspace));
  memset(result->streams,0,nb_streams*sizeof(struct stream_workspace));
  for(i=0;i<nb_streams;i++) {
    result->streams[i].cd_launched=0;
    result->streams[i].id=fid+i;
    result->streams[i].output=-1;
    result->streams[i].socket=NULL;
    result->streams[i].shdata=NULL;
    result->streams[i].acqunit=result;
    result->streams[i].name=malloc(1024);
    sprintf(result->streams[i].name,"%s_%d",name,i);
    printf("creating stream %d with id %d and name %s\n",i,result->streams[i].id,result->streams[i].name);
  }
  
  //open new files
  debug("open the new files\n");
  if (opennew_files(result)<0) {
    return NULL;
  }
  
  //open new sockets
  debug("open the new sockets\n");
  opennew_socket(result);
  
  //open new shared buffers
  debug("open the new shared buffers\n");
  opennew_shmem(result,timeout);

  usleep(1000);

  return result;
}

void del_acq_unit(struct acqunit_workspace * auws) {
  struct acqunit_workspace *au;
  struct acq_workspace *ws=auws->globalws;
  struct data_descriptor *current_ctrl;
  int port;
  int i;
  struct cmd_result *cres;

  //stop and deinit all acqunit
  printf("send exit order to acqunit thread\n");
  //stop the aquisition if necessary
  if (auws->acq_active) {
    auws->acq_active=0;
    usleep(50000);
  }

  //deinit order
  auws->deinitialized=0;
  auws->deinit=1;

  //wait for threads to exit
  printf("waiting for acq thread to exit\n");
  while(auws->deinitialized==0) {
    usleep(5000);
  }

   //close the files
  debug("close the files\n");
  close_files(auws);
  free(auws->file_mutex);

  //close the sockets
  debug("close the sockets\n");
  close_socket(auws);

  //close the shared memory
  debug("close the shared memory\n");
  close_shmem(auws);

  //kill the converters if necessary
  for(i=0;i<auws->nb_streams;i++) {
    if (auws->streams[i].cd_launched) {
      port=submod_getport("CD_BASE_PORT")+i;
      printf("killing converter %d on port %d\n",i,port);
      cres=sendcmd("localhost",port,"exit_module","end");
      free_cmd_result(cres);
    }
    free(auws->streams[i].name);
  }

  // unlink any control packet from deleted acq unit
  current_ctrl=ws->begin_ctrl;
  while(current_ctrl!=ws->end_ctrl) {
    if (current_ctrl->id==-1) {
      current_ctrl=current_ctrl->next;
      continue;
    }
    if (current_ctrl->auws==auws) {
      current_ctrl->auws=NULL;
    }
    current_ctrl=current_ctrl->next;
  }
  
  //unchain the ws
  if (ws->acqunits==auws) {
    ws->acqunits=auws->next;
  } else {
    au=ws->acqunits;
    while(au!=NULL) {
      if (au->next==auws)
        au->next=auws->next;
      au=au->next;
    }
  }

  //free the structures
  dlclose(auws->lib_acq);
  dlclose(auws->lib_uncap);
  free(auws->acq_active_mutex);
  free(auws->streams);
  free(auws->datadir);
  free(auws->convert_script);
  free(auws->name);
  free(auws);
}

void start_acquisition(struct acqunit_workspace *auws) {
  pthread_mutex_lock(auws->acq_active_mutex);
  auws->acq_active=1;
  pthread_mutex_unlock(auws->acq_active_mutex);
}


void stop_acquisition(struct acqunit_workspace *auws) {
  pthread_mutex_lock(auws->acq_active_mutex);
  auws->acq_active=0;
  pthread_mutex_unlock(auws->acq_active_mutex);
}

int get_acqstate(struct acqunit_workspace *auws) {
  int active;
  pthread_mutex_lock(auws->acq_active_mutex);
  active=auws->acq_active;
  pthread_mutex_unlock(auws->acq_active_mutex);
  return active;
}

//get an acquisition unit structure by its name
struct acqunit_workspace *get_acqunit(struct acq_workspace *ws,char *name) {
  struct acqunit_workspace *iter=ws->acqunits;
  while(iter!=NULL) {
    if (!strcmp(iter->name,name))
      return iter;
    iter=iter->next;
  }
  return NULL;
}
