/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include "libacq.h"

void *init() {
  char *varmod_ip;
  
  //common workspace declarations
  struct acq_workspace *ws=malloc(sizeof(struct acq_workspace));
  memset(ws,0,sizeof(struct acq_workspace));

  //common workspace initialization
  ws->deinit=0;
  ws->bsched_active=0;
  ws->data_port_base=submod_getport("DATASTREAM_BASE_PORT");

  //initializing the acquisition units to 0
  ws->nb_streams=0;
  ws->acqunits=NULL;

  //buffers pool init
  ws->bpool=init_buffer_pool(NB_GENERIC_BUFFERS);

  //data descriptors pool init
  ws->init_data=init_data_descriptor_list(NB_DATA_DD);
  ws->current_data=ws->init_data;
  ws->data_descriptor_mutex=malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(ws->data_descriptor_mutex,NULL);

  //control descriptors pool init
  ws->init_ctrl=init_data_descriptor_list(NB_CTRL_DD);
  ws->end_ctrl=ws->init_ctrl;
  ws->begin_ctrl=ws->init_ctrl;
  ws->ctrl_descriptor_mutex=malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(ws->ctrl_descriptor_mutex,NULL);

  //creating the netserv
  ws->nw=start_empty_netserv(500);
  
  
  ws->t_burst_detector=(pthread_t*)malloc(sizeof(pthread_t));
  if (pthread_create(ws->t_burst_detector,NULL,burst_detector,ws)<0) {
    perror("pthread_create");
    exit(1);
  }
  pthread_detach(*ws->t_burst_detector);

  //init consumers chain variables
  ws->deinit_mutex=malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(ws->deinit_mutex,NULL);
  pthread_mutex_unlock(ws->deinit_mutex);
  
  //initialize stats system
  varmod_ip=submod_getip("getvar@varmod",-1);
  ws->stats=init_stats(varmod_ip,1000000);
  free(varmod_ip);
  add_acc(ws->stats,"nb_data_pkts",0);
  add_acc(ws->stats,"nb_lost_pkts",0);
  add_acc(ws->stats,"nb_corr_pkts",0);
  add_acc(ws->stats,"nb_ctrl_pkts",0);
  add_acc(ws->stats,"bytes_on_dsk",0);
  add_acc(ws->stats,"bytes_on_skt",0);
  add_acc(ws->stats,"bytes_on_shm",0);

  //initializing the acquisition
  init_acquisition(ws);
  
  //initializing the scheduler
  init_burst_detector(ws);
  
  //initializing all the other modules
  launch_consumer_chain(ws);
  
  return ws;
}

//pyrapi:start_converter_acq:acq_id,stream_id,plugin_path
//other params : plugin params
struct cmd_result * start_converter_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  
  int stream_id;
  char *acq_id;
  char *path;
  struct incr_string *cmd=new_is();
  struct acqunit_workspace *auws;
  int shmem_id;
  int i;
  int syst_res;
  int socket_port;
  int ok;
  int tst_sock;
  int cmd_port;

  //check the number of params
  if (command->nb_params<3) {
    return build_result(0,"start_converter_acq, %d arguments given, at least 3 needed : acq_id stream_id converter_plugin",command->nb_params);
  }
  
  //get the parameters
  acq_id=command->params[0];
  stream_id=atoi(command->params[1]);
  path=command->params[2];

  //find the acqunit
  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"unknown acq_id %s",acq_id);
  }

  //verify that the stream exists
  if(stream_id>=auws->nb_streams) {
    return build_result(0,"unknown stream id %s in acq %s",stream_id,acq_id);
  }

  //verify that the converter has not been launched yet
  if(auws->streams[stream_id].cd_launched==1) {
    return build_result(0,"converter already launched");
  }

  //launching the converter
  shmem_id=stream_id+auws->fid;
  socket_port=submod_getport("DATASTREAM_BASE_PORT")+stream_id+auws->fid;
  cmd_port=submod_getport("CD_BASE_PORT")+stream_id+auws->fid;
  debug("launching converter for stream %d in acq %s on port %d shmem_id %d or tcp port %d\n",stream_id,acq_id,cmd_port,shmem_id,socket_port);

  //shmem
  fsuffix_is(cmd,"/usr/bin/cmdmod /opt/pyrame/cmd_converter.xml -a %s %d shmem://%d %s",auws->streams[stream_id].name,cmd_port,shmem_id,path);
  
  //socket
  //fsuffix_is(cmd,"/usr/bin/cmdmod /opt/pyrame/cmd_converter.xml -a %s %d tcp://localhost:%d %s",auws->streams[stream_id].name,cmd_port,socket_port,path);

  //valgrind
  //fsuffix_is(cmd,"xfce4-terminal -H -x valgrind --leak-check=full /usr/bin/cmdmod /opt/pyrame/cmd_converter.xml -d -a %s %d shmem://%d %s",auws->streams[stream_id].name,cmd_port,shmem_id,path);
  //fsuffix_is(cmd,"/bin/gnome-terminal -x valgrind --leak-check=full --show-leak-kinds=all /usr/bin/cmdmod /opt/pyrame/cmd_converter.xml -a %s %d tcp://localhost:%d %s",auws->streams[stream_id].name,cmd_port,socket_port,path);

  //gdb
  //fsuffix_is(cmd,"xfce4-terminal -H -x gdb -ex run --args /usr/bin/cmdmod /opt/pyrame/cmd_converter.xml -d -a %s %d shmem://%d '%s'",auws->streams[stream_id].name,cmd_port,shmem_id,path);
  //fsuffix_is(cmd,"/bin/gnome-terminal -x gdb -ex run --args /usr/bin/cmdmod /opt/pyrame/cmd_converter.xml -d -a %s %d shmem://%d '%s'",auws->streams[stream_id].name,cmd_port,shmem_id,path);
  
  
  for(i=3;i<command->nb_params;i++) {
    fsuffix_is(cmd," '%s'",command->params[i]);
  }
  fsuffix_is(cmd," &");
  printf("system command : %s\n",cmd->str);
  syst_res=system(cmd->str);
  if (syst_res!=0) {
    printf("system error : %d\n",syst_res);
  }

  //check if converter is listening on command socket
  ok=0;
  //printf("checking converter opening on port %d\n",cmd_port);
  for(i=1;i<10;i++) {
    usleep(5000*i);
    //printf("open socket\n");
    tst_sock=open_socket("localhost",cmd_port);
    if (tst_sock==-1) {
      //printf("nope\n");
    } else {
      //printf("ok\n");
      ok=1;
      close(tst_sock);
      break;
    }
  }

  free_is(cmd);

  if (!ok)
    return build_result(0,"unable to launch converter on port %d",cmd_port);

  //mark the converter as launched for the stream
  auws->streams[stream_id].cd_launched=1;
  return build_result(1,"ok");
  
}

//pyrapi:deinit_acq:acq_id
struct cmd_result * deinit_acq(struct cmd *command,void *workspace) {
  //param 0 : acqunit_name
  char *acq_id;
  struct acqunit_workspace * auws;
  struct acq_workspace *ws=(struct acq_workspace *)workspace;

  //get parameters
  if (command->nb_params!=1) {
    return build_result(0,"deinit_acq, %d arguments given, 1 needed",command->nb_params);
  }

  //get parameters
  acq_id=command->params[0];

  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"no acqunit with that name (%s)",acq_id);
  }

  del_acq_unit(auws);
  return build_result(1,"ok");
}

//pyrapi:init_acq:acq_id,nb_streams,acq_lib_name,uncap_lib_name,acq_plugin_param1,acq_plugin_param2,acq_plugin_param3,shmem_timeout,replay_enabled,datadir,convert_script
struct cmd_result * init_acq(struct cmd *command,void *workspace) {

  struct acqunit_workspace * newunit;
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  char *acq_id;
  int nbs;
  int fid;
  char *acq_lib, *uncap_lib;
  char *acq_plugin_param1, *acq_plugin_param2, *acq_plugin_param3;
  int timeout,replay;
  char *datadir, *convert_script;

  //get parameters
  if (command->nb_params!=11) {
    return build_result(0,"init_acq, %d arguments given, 11 needed",command->nb_params);
  }

  //get parameters
  //param 0 : acq_id
  //param 1 : nb_streams
  //param 2 : name of the acq library
  //param 3 : name of the uncap library
  //param 4 : acq plugin param 1
  //param 5 : acq plugin param 2
  //param 6 : acq plugin param 3
  //param 7 : timeout for shared memory in us
  //param 8 : replay enabled
  //param 9 : datadir
  //param 10 : convert script
  acq_id=command->params[0];
  nbs=iohtoi(command->params[1]);
  acq_lib=command->params[2];
  uncap_lib=command->params[3];
  acq_plugin_param1=command->params[4];
  acq_plugin_param2=command->params[5];
  acq_plugin_param3=command->params[6];
  timeout=iohtoi(command->params[7]);
  replay=iohtoi(command->params[8]);
  datadir=command->params[9];
  convert_script=command->params[10];

  newunit=get_acqunit(ws,acq_id);
  if (newunit!=NULL) {
    return build_result(0,"an acqunit with that name (%s) already exists",acq_id);
  }

  //create the workspace for the new unit
  printf("creating newunit workspace\n");
  fid=ws->nb_streams;
  debug("nbs=%d fid=%d\n",nbs,fid);
  newunit=make_acq_unit(acq_id,nbs,fid,acq_lib,uncap_lib,datadir,convert_script,timeout,replay,ws);
  if (newunit==NULL) {
    return build_result(0,"error creating acqunit");
  }
  
  //initialize acquisition library
  printf("initializing acquisition library\n");
  newunit->specws=(newunit->init_acq)(acq_plugin_param1,acq_plugin_param2,acq_plugin_param3);

  //initialize uncap library
  printf("initializing uncap library\n");
  newunit->uncapws=(newunit->init_uncap)(nbs,fid,acq_id);

  //initialize the nb of streams 
  ws->nb_streams+=nbs;
  
  //store the newunit in the main ws
  printf("linking the new unit\n");
  newunit->next=ws->acqunits;
  ws->acqunits=newunit;

  //start the thread
  printf("Starting the acquisition thread\n");
  if (pthread_create(&(newunit->t_acquisition),NULL,acquisition,newunit)<0) {
    perror("pthread_create");
    exit(1);
  }
  pthread_detach(newunit->t_acquisition);

  //return the acqunit id as a result
  return build_result(1,"ok");
}

//pyrapi:start_acq:acq_id
struct cmd_result * start_acq(struct cmd *command,void *workspace) {

  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  int i;
  struct cmd_result *cres;
  int port;
  struct acqunit_workspace *auws;
  char *acq_id;

 if (command->nb_params!=1) {
    return build_result(0,"start_acq, %d arguments given, 1 needed",command->nb_params);
  }

  acq_id=command->params[0];

  //get the workspace of the acqunit
  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"unknown acqunit %s",acq_id);
  }
  

  //start the acquisition on every related converters
  for(i=0;i<auws->nb_streams;i++) {
    if (auws->streams[i].cd_launched) {
      port=submod_getport("CD_BASE_PORT")+i+auws->fid;
      debug("start_acq converter %d on %s\n",i,auws->name);
      cres=sendcmd("localhost",port,"start_acq_converter","end");
      if (cres->retcode==0) {
	return build_result(0,"cant start acq on converter %d on auws %s\n",i,auws->name);
      }
      free_cmd_result(cres);
    }
  }

  //starting acquisition
  start_acquisition(auws);
  return build_result(1,"ok");
}

//pyrapi:stop_acq:acq_id
struct cmd_result * stop_acq(struct cmd *command,void *workspace) {

  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  int i;
  struct cmd_result *cres;
  int port;
  struct acqunit_workspace *auws;
  char *acq_id;

  if (command->nb_params!=1) {
    return build_result(0,"start_acq, %d arguments given, 1 needed",command->nb_params);
  }

  acq_id=command->params[0];

  //get the workspace of the acqunit
  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"unknown acqunit %s",acq_id);
  }

  printf("stoping acquisition\n");
  stop_acquisition(auws);
  usleep(500000);

  for(i=0;i<auws->nb_streams;i++) {
    if (auws->streams[i].cd_launched) {
      port=submod_getport("CD_BASE_PORT")+i+auws->fid;
      cres=sendcmd("localhost",port,"stop_acq_converter","end");
      free_cmd_result(cres);
    }
  }

  return build_result(1,"ok");
}

//pyrapi:flush_files_acq:acq_id,acq_name,script_param1,script_param2,script_param3,script_param4,script_param5,script_param6
struct cmd_result * flush_files_acq(struct cmd *command,void *workspace) {
  char *acq_id;
  char *acq_name;
  char *script_param1,*script_param2,*script_param3,*script_param4,*script_param5,*script_param6;
  struct acqunit_workspace *auws;

  struct acq_workspace *ws=(struct acq_workspace *)workspace;

  //check the number of parameters
  if (command->nb_params<8) {
    return build_result(0,"flush_files_acq, %d arguments given, 8 needed",command->nb_params);
  }

  acq_id=command->params[0];
  acq_name=command->params[1];
  if (!strcmp(acq_name,"")) {
    printf("warning : no acq_name given, using noname as prefix\n");
    acq_name="noname";
  }
  script_param1=command->params[2];
  script_param2=command->params[3];
  script_param3=command->params[4];
  script_param4=command->params[5];
  script_param5=command->params[6];
  script_param6=command->params[7];


  //get the workspace of the acqunit
  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"unknown acqunit %s",acq_id);
  }
  
  printf("flushing files\n");
  if (get_acqstate(auws)!=0) {
    return build_result(0,"acquisition unit is still acquiring, cant flush");
  }
  
  flush_files(auws,acq_name,script_param1,script_param2,script_param3,script_param4,script_param5,script_param6);

  flush_shmem(auws);

  return raw_build_result(1,"ok");
}

//pyrapi:get_stats_acq:
struct cmd_result * get_stats_acq(struct cmd *command,void *workspace) {
  char *res;
  struct cmd_result *pres;
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  res=get_stats(ws->stats);
  pres=build_result(1,res);
  free(res);
  return pres;
}


//pyrapi:zero_stats_acq:
struct cmd_result * zero_stats_acq(struct cmd *command,void *workspace) {
  int i;
  int port;
  struct cmd_result *cres;
  struct acqunit_workspace *auws;

  struct acq_workspace *ws=(struct acq_workspace *)workspace;

  //zero the stats of the acq chain
  zero_stats(ws->stats);

  auws=ws->acqunits;
  while(auws!=NULL) {

    //zero the stats of all the converter/dispatchers
    for(i=0;i<auws->nb_streams;i++) {
      if (auws->streams[i].cd_launched) {
        port=submod_getport("CD_BASE_PORT")+i+auws->fid;
        cres=sendcmd("localhost",port,"reset_stats_converter","end");
        free_cmd_result(cres);
      }
    }
    auws=auws->next;
  }

  return build_result(1,"ok");
}

//pyrapi:get_cpkt_byid_acq:acq_id,id1,id2,id3,id4
struct cmd_result * get_cpkt_byid_acq(struct cmd *command,void *workspace) {

  int id1,id2,id3,id4;
  int i;
  struct data_descriptor *current_ctrl;
  char * encoded;
  struct cmd_result *pres;
  char *acq_id;
  struct acq_workspace *ws=(struct acq_workspace *)workspace;

  //check the number of parameters
  if (command->nb_params<5) {
    return build_result(0,"get_cpkt_byid_acq, %d arguments given, 5 needed",command->nb_params);
  }

  //extract the parameters
  acq_id=command->params[0]; 
  id1=iohtoi(command->params[1]); //this is id1 of uncap plugin
  id2=iohtoi(command->params[2]); //this is id2 of uncap plugin
  id3=iohtoi(command->params[3]); //this is id3 of uncap plugin
  id4=iohtoi(command->params[4]); //this is id4 of uncap plugin
  debug("searching for %s %x %x %x %x :\n",acq_id,id1,id2,id3,id4);
    
  for(i=1;i<10;i++) { //we try 10 times
    //wait for packet to come - the more chess we have the longer we wait
    usleep(i*50000);

    //search the packet in the ctrl fifo
    current_ctrl=ws->begin_ctrl;
    while(current_ctrl!=ws->end_ctrl) {
      if (current_ctrl->id==-1) {
        current_ctrl=current_ctrl->next;
        continue;
      }
      if (current_ctrl->auws && !strcmp(acq_id,current_ctrl->auws->name)) //the acq_id is good
        if ((current_ctrl->auws->select_packet_uncap)(current_ctrl->auws->uncapws,current_ctrl->generic_buffer->data,id1,id2,id3,id4)) {
          //result found, encode and send back
          encoded=encode_packet(current_ctrl->generic_buffer->data,current_ctrl->generic_buffer->size);
          pres=build_result(1,encoded);
          free(encoded);
          //remove the found packet
          current_ctrl->id=-1;
          return pres;
        } //if found
      current_ctrl=current_ctrl->next;
    }
  }
  //Too much failure, we give up
  return build_result(0,"no packet with this id");
  }


//pyrapi:start_bsched_acq:
struct cmd_result * start_bsched_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  ws->bsched_active=1;
  return build_result(1,"ok");
}

//pyrapi:stop_bsched_acq:
struct cmd_result * stop_bsched_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  ws->bsched_active=0;
  return build_result(1,"ok");
}

//pyrapi:allow_autoflush_acq:
struct cmd_result * allow_autoflush_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  
  if (command->nb_params<1) {
    return build_result(0,"allow_autoflush_acq, %d arguments given, 1 needed",command->nb_params);
  }

  ws->autoflush_limit=iohtoi(command->params[0]);
  ws->autoflush_active=1;
  return build_result(1,"ok");
}

//pyrapi:dis_autoflush_acq:
struct cmd_result * dis_autoflush_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  ws->autoflush_active=0;
  return build_result(1,"ok");
}

//pyrapi:inject_data_acq:acq_id,stream_id,data
struct cmd_result * inject_data_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  int streamid;
  char *acq_id;
  char *data;
  struct generic_buffer *acq_buffer;
  unsigned char *binary_data;
  struct acqunit_workspace *auws;

  if (command->nb_params<3) {
    return build_result(0,"inject_data_acq, %d arguments given, 3 needed",command->nb_params);
  }

  //get the parameters
  acq_id=command->params[0];
  streamid=iohtoi(command->params[1]);
  data=command->params[2];

  //get the workspace of the acqunit
  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"unknown acqunit %s",acq_id);
  }
    
  //check that stream exists
  if (streamid>=auws->nb_streams) {
    return build_result(0,"no stream with that id");
  }

    //create a new buffer
    acq_buffer=get_generic_buffer(ws->bpool);

    //copy the data
    binary_data=decode_binary_data(data,&acq_buffer->size);
    strncpy((char *)acq_buffer->data,(const char *)binary_data,acq_buffer->size);
    free(binary_data);
    
    //fill a data descriptor with this packet
    insert_new_data_descriptor(ws,acq_buffer,auws,0);
    
    //return result
    return build_result(1,"data injected");
}

//pyrapi:send_data_acq:acq_id,stream_id,data,addressing_param1,addressing_param2,addressing_param3
struct cmd_result * send_data_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  struct cmd_result *cres;
  int streamid;
  int size;
  char *data;
  char *acq_id;
  unsigned char *binary_data;
  struct acqunit_workspace *auws;

  if (command->nb_params<6) {
    return build_result(0,"send_data_acq, %d arguments given, 6 needed",command->nb_params);
  }
  
  //get parameters
  acq_id=command->params[0];
  streamid=iohtoi(command->params[1]);
  data=command->params[2];
  
  //get the workspace of the acqunit
  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"unknown acqunit %s",acq_id);
  }
  
  if (streamid>=auws->nb_streams) {
    return build_result(0,"no stream with that id");
  }

  //check that send_acq exists for this plugin
  if (auws->send_acq==NULL) {
    return build_result(0,"no send_acq function implemented for this module\n");
  }
    
  //decode the data
  binary_data=decode_binary_data(data,&size);

  //send the data
  cres=(auws->send_acq)(auws->specws,binary_data,size,command->params+3,command->nb_params-3);

  //free the binary data block
  free(binary_data);

  return cres;
}
  
//pyrapi:get_ports_acq:acq_id
struct cmd_result * get_ports_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  int j;
  struct incr_string *ires=new_is();
  int port;
  char *acq_id;
  int first_res=1;
  struct acqunit_workspace *auws;

  if (command->nb_params<1) {
    return build_result(0,"get_ports_acq, %d arguments given, 1 needed",command->nb_params);
  }
  acq_id=command->params[0];

  //get the workspace of the acqunit
  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"unknown acqunit %s",acq_id);
  }

  for (j=0;j<auws->nb_streams;j++) {
    port=auws->streams[j].id+ws->data_port_base;
    if (first_res) {
      first_res=0;
      fsuffix_is(ires,"%d",port);
    } else {
      fsuffix_is(ires,",%d",port);
    }    
  }
  
  return build_result(1,reduce_is(ires));
}

//pyrapi:set_stream_name_acq:acq_id,stream_id,stream_name
struct cmd_result * set_stream_name_acq(struct cmd *command,void *workspace) {
  struct acq_workspace *ws=(struct acq_workspace *)workspace;
  int stream_id;
  struct acqunit_workspace *auws;
  char *acq_id, *stream_name;

  if (command->nb_params!=3) {
    return build_result(0,"set_stream_name_acq, %d arguments given, 3 needed",command->nb_params);
  }

  acq_id=command->params[0];
  stream_id=atoi(command->params[1]);
  stream_name=command->params[2];

  //get the workspace of the acqunit
  auws=get_acqunit(ws,acq_id);
  if (auws==NULL) {
    return build_result(0,"unknown acqunit %s",acq_id);
  }

  //check stream id
  if (stream_id>=auws->nb_streams) {
    return build_result(0,"no stream with this id");
  }
  
  free(auws->streams[stream_id].name);
  auws->streams[stream_id].name=strdup(stream_name);
  return build_result(1,"ok");
}

  

