/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <net/if.h>
#include <fcntl.h>
#include <math.h>
#include <expat.h>
#include <dlfcn.h>

#include <cmdmod_helper.h>

#ifndef LIBACQ_H
#define LIBACQ_H

#define ETH_P_ALL	0x0003
#define BACKLOG 100
//#define NETMAXCLIENTS 100
#define MAXID 4

//this is the global workspace of the whole acquisition chain
struct acq_workspace {
  int nb_streams; 
  //acquisition units chained list
  struct acqunit_workspace *acqunits;
  //reference clock
  char *refclock;
  //scheduler
  pthread_t *t_burst_detector;
  int bsched_active; //the scheduler is active or not
  unsigned char burst_state; //are we in a burst or not?
  pthread_mutex_t *burst_state_mutex;
  pthread_mutex_t **sched_clients;
  int sched_nb_clients;
  unsigned char packet_detector;
  //data descriptors
  struct data_descriptor *init_data;
  struct data_descriptor *current_data;
  pthread_mutex_t *data_descriptor_mutex;
  struct data_descriptor *init_ctrl;
  struct data_descriptor *begin_ctrl;
  struct data_descriptor *end_ctrl;
  pthread_mutex_t *ctrl_descriptor_mutex;
  struct buffer_pool *bpool;
  
  //netserv
  struct netserv_workspace *nw;
  //sockets
  struct netserv_client **sockets_nc;
  //struct ports_table *ports;
  unsigned short data_port_base;
  //shared memory
  int active_shmem;
  
  //consumers
  int deinit;
  pthread_mutex_t *deinit_mutex;
  int endinit;
  struct consumer_workspace **consumers;

  int autoflush_active; //flush automatically the files when autoflush_limit is reached
  int autoflush_limit; //in Mo
  int rawsize;

  //stats
  struct stats_ws *stats;

} acq_workspace;


struct stream_workspace {
  //output file
  int output;
  //output socket
  struct netserv_client *socket;
  //shared memory
  struct shbuffer *shdata;
  //stream id
  int id;
  char *name;
  //corresponding acquisition unit
  struct acqunit_workspace *acqunit;
  //converter dispatcher
  int cd_launched;
};

//this is the specific workspace for every acquisition unit
struct acqunit_workspace {
  char *name;
  //streams
  int nb_streams;
  int fid;
  struct stream_workspace *streams;
  //int first_stream; // the id of the first stream of this acquisition unit
  //int last_stream;
  int deinit; //put it to 1 to make the acqunit to stop its thread
  int deinitialized; // set to 1 when the acq and uncap of this au is deinit
  //acquisition library
  void *lib_acq;
  void * (*init_acq)();
  void (*deinit_acq)();
  void (*start_acq)();
  void (*stop_acq)();
  struct cmd_result *(*send_acq)();
  int (*acquire_one_block_acq)();
  //uncap library
  void *lib_uncap;
  void * (*init_uncap)();
  void (*deinit_uncap)();
  int (*prefilter_uncap)();
  int (*uncap_uncap)();
  int (*select_packet_uncap)(); 
  //acquistion plugin workspace
  void *specws;
  //uncap plugin workspace
  void *uncapws;
  //global workspace
  struct acq_workspace *globalws;
  //acquisition thread
  pthread_t t_acquisition;
  //data directory
  char *datadir;
  //data conversion
  char *convert_script;
   pthread_mutex_t *file_mutex;
  //threads synchronization
  unsigned char acq_active; //1 if acquisition is active
  pthread_mutex_t *acq_active_mutex;
  //replay
  char replay_enabled;
  int replay_file;
  //shmem
  int timeout;
  //chaining 
  struct acqunit_workspace *next;
} acqunit_workspace;

//burst detector
#define OUTOFBURST 0
#define INBURST 1
#define BURST_TRESHOLD 2000 //microseconds

struct shared_cmds {
  unsigned char flush_requested;
  int flush_result;
  char *flush_prefix;
  pthread_mutex_t *flush_synchro;
} shared_cmds;

//Number of data descriptor for data and control
#define NB_DATA_DD 100000
#define NB_CTRL_DD 4096

//specific parts headers

//buffers
//128 kbuffers, 4k by buffer, 512 Mo
#define NB_GENERIC_BUFFERS 131072
#define MAX_PACKET_SIZE 4096

struct data_descriptor {
  //set by acquisition for debugging
  int acqid;

  //structure linking
  struct data_descriptor *next, *prev;
  
  //data container
  struct generic_buffer * generic_buffer;

  //acquisition unit 
  struct acqunit_workspace *auws;

  //uncap fields
  unsigned char *data;
  int data_size;
  int streamid;

  //the following id is used for consumers to know if it is their turn to treat
  //-1 means free
  int id;
  int corrupted;

} data_descriptor;

struct buffer_pool {
  struct generic_buffer *generic_buffer_head;
  pthread_mutex_t generic_buffers_mutex;
} buffer_pool;

struct data_buffer {
  struct data_buffer *next, *prev;
  unsigned char free;
  int size;
  unsigned char data[MAX_PACKET_SIZE];
} data_buffer;


struct generic_buffer {
  struct generic_buffer *next;
  int size;
  unsigned char data[MAX_PACKET_SIZE];
} generic_buffer;


struct buffer_pool *init_buffer_pool(int);
struct generic_buffer *init_generic_buffers(int);
struct generic_buffer *get_generic_buffer(struct buffer_pool * );
void leave_generic_buffer(struct buffer_pool * ,struct generic_buffer *);
struct data_descriptor* init_data_descriptor_list(int);

char encode_quartet(char c);
char *encode_packet(unsigned char *,int);
unsigned char *decode_binary_data(char *,int *);

//consumers
struct consumer_workspace {
  struct acq_workspace *gws;
  int myid;
  void (*myinit)();
  void (*mydeinit)();
  int (*mytreatment)();
  struct data_descriptor *init_descriptor;
  char *name;
  pthread_t mythread;
  int cons_pipe[2];
} consumer_workspace;


void *generic_consumer(void *);
void destructor_init(struct consumer_workspace *);
void destructor_deinit(struct consumer_workspace *);
int destructor_treatment(struct data_descriptor *,struct consumer_workspace *);
void launch_consumer_chain(struct acq_workspace *);
void stop_consumer_chain(struct acq_workspace *);

//acquisition
void *acquisition(void *);
void init_acquisition(struct acq_workspace *ws);
void deinit_acquisition(struct acq_workspace *);
int insert_new_data_descriptor(struct acq_workspace *,struct generic_buffer *,struct acqunit_workspace *,int);
int insert_new_ctrl_descriptor(struct acq_workspace *ws,struct generic_buffer *acq_buffer,struct acqunit_workspace *auws);

//acquisition units
struct acqunit_workspace * make_acq_unit(char *name,int nb_streams,int fid,char * acq_lib,char * uncap_lib,char *datadir,char *convert_script,int timeout,int replay,struct acq_workspace *ws);
void del_acq_unit(struct acqunit_workspace * auws);
//void free_acq_unit(struct acqunit_workspace *auws);
void start_acquisition(struct acqunit_workspace *auws);
void stop_acquisition(struct acqunit_workspace *auws);
int get_acqstate(struct acqunit_workspace *auws);
struct acqunit_workspace *get_acqunit(struct acq_workspace *ws,char *name);

//data socket
void data_socket_init(struct consumer_workspace *);
void data_socket_deinit(struct consumer_workspace *);
int data_socket_treatment(struct data_descriptor *,struct consumer_workspace *);
void opennew_socket(struct acqunit_workspace * auws);
void close_socket(struct acqunit_workspace * auws);

//cache
void cache_init(struct consumer_workspace *);
void cache_deinit(struct consumer_workspace *);
int cache_treatment(struct data_descriptor *,struct consumer_workspace *);
int opennew_files(struct acqunit_workspace *auws);
void open_replay_file(struct acqunit_workspace *auws);
void close_replay_file(struct acqunit_workspace *auws);
void close_files(struct acqunit_workspace *auws);
void flush_files(struct acqunit_workspace *auws,char *prefix, char *param1, char *param2, char *param3, char *param4,char *param5,char *param6);

//scheduler
void * burst_detector(void *);
void add_sched_client(struct acq_workspace *,pthread_mutex_t *);
void init_burst_detector(struct acq_workspace *);
void deinit_burst_detector(struct acq_workspace *);

//uncap
#define PREF_JUNK 0
#define PREF_DATA 1
#define PREF_CTRL 2

void uncap_init(struct consumer_workspace *);
void uncap_deinit(struct consumer_workspace *);
int uncap_treatment(struct data_descriptor *,struct consumer_workspace *);

//shmem
void shmem_init(struct consumer_workspace *ws);
void flush_shmem(struct acqunit_workspace * auws);
void opennew_shmem(struct acqunit_workspace * auws,int timeout);
void close_shmem(struct acqunit_workspace * auws);
void shmem_deinit(struct consumer_workspace *ws);
int shmem_treatment(struct data_descriptor *ad,struct consumer_workspace *ws);

#endif
