#!/usr/bin/env python2
# -*- coding: utf-8 -*-
# 
# Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
# This file is part of Pyrame.
# 
# Pyrame is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Pyrame is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrame.  If not, see <http://www.gnu.org/licenses/>

import time,pools

DEBUG = 0
if (DEBUG): import sys

def unescape_string(string):
    #return string.replace(r"\e",chr(27)).replace(r"\t",chr(9)).replace(r"\n",chr(10)).replace(r"\r",chr(13))
    return string.decode('string_escape')

def decode_bin_string(string):
    if len(string)%2!=0:
        raise Exception("data must have a length multiple of 2. Current length: %d"%(len(string)))
    result=""
    for i in range(len(string)/2):
        result += chr(int(string[2*i:2*i+2],16))
    return result

def encode_bin_string(string):
    result=""
    for c in string:
        result += "%02x"%(ord(c))
    return result

def ignore_chars(string):
    return string.replace("\0","").replace("\n","").replace("\r","")

class buses(object):
    def __init__(self):
        self.ll_link_pool=pools.pool("ll_link")
        self.ll_links={}
    # PURE VIRTUAL METHODS

    def init(self,link_id,conf_string):
        raise NotImplementedError

    def config(self,link_id):
        raise NotImplementedError

    def ll_send(self,ll_link,link,data):
        raise NotImplementedError

    def ll_receive(self,ll_link,link,bytes_to_read,timeout="undef"):
        raise NotImplementedError

    def ll_open(self,link):
        raise NotImplementedError

    def ll_close(self,link):
        raise NotImplementedError
        
    def ll_new_link(self,link):
        try:
            ll_link=self.ll_link_pool.get(link["descr"])
        except:
            self.ll_link_pool.new(link["descr"],{"counter":0})
            ll_link=self.ll_link_pool.get(link["descr"])
        ll_link["counter"]+=1
        retcode,res=self.ll_connect(link)
        if retcode==1:
            link["connected"]=1
        return retcode,res

    def ll_del_link(self,link):
        try:
            ll_link=self.ll_link_pool.get(link["descr"])
        except:
            return
        ll_link["counter"]-=1
        if ll_link["counter"]<1:
            self.ll_close(self.ll_links[link["descr"]])
            del self.ll_links[link["descr"]]
            self.ll_link_pool.remove(link["descr"])
        link["connected"]=0

    def ll_connect(self,link):
        if link["descr"] not in self.ll_links:
            ll=self.ll_open(link)
            if ll==None:
                return 0,"unable to connect link to %s"%(link["descr"])
            self.ll_links[link["descr"]]=ll
        return 1,"ok"

    # METHODS
    def write(self,link,mode,data):
        retcode,res=self.ll_connect(link)
        if retcode==0:
            return 0,res
        if mode=="escaped":
            data = unescape_string(data)
        elif mode=="bin":
            try:
                data = decode_bin_string(data)
            except Exception as e:
                return 0,str(e)
        else:
            return 0,"Wrong write mode: %s" % (mode)
        # Send data
        retcode,res = self.ll_send(self.ll_links[link["descr"]],link,data)
        if retcode == 0:
            return 0,res
        return 1,"ok"

    def read(self,link,mode,bytes_to_read,timeout="undef"):
        retcode,res=self.ll_connect(link)
        if retcode==0:
            return 0,res
        retcode,res = self.ll_receive(self.ll_links[link["descr"]],link,bytes_to_read,timeout)
        if retcode == 0:
            return 0,res
        if mode=="ignore_chars":
            res = ignore_chars(res)
        if mode=="bin":
            res = encode_bin_string(res)
        return 1,res

    def read_until(self,link,mode,eot,timeout="undef"):
        retcode,res=self.ll_connect(link)
        if retcode==0:
            return 0,res
        eot_tab=[]
        if eot=="" or eot=="undef":
            return 0,"End Of Transmission (eot) parameter needed"
        for c in eot.split(","):
            eot_tab.append(unescape_string(c))
        # Set timeout
        oldtime = time.time()
        if timeout=="" or timeout=="undef":
            timeout = float(link["timeout"])
        else:
            timeout = float(timeout)
        # Read until EOT
        response = ""
        found = [False]*len(eot_tab)
        while not any(found) and timeout>0:
            retcode,res = self.ll_receive(self.ll_links[link["descr"]],link,1,timeout)
            if retcode==0:
                return 0,res
            response += res
            timeout -= time.time() - oldtime
            oldtime = time.time()
            found = list((response[-len(c):] in eot_tab) for c in eot_tab)
        if not any(found) and timeout<=0:
            return 0,"Timeout waiting for EOT character"
        detected_eot = eot_tab[found.index(True)]
        response = response[0:-len(detected_eot)]
        if mode=="ignore_chars":
            response = ignore_chars(response)
        if mode=="bin":
            response = encode_bin_string(response)
        return 1,response

    def expect(self,link,pattern,timeout="undef"):
        retcode,res=self.ll_connect(link)
        if retcode==0:
            return 0,res
        # Unescape pattern
        pattern = unescape_string(pattern)
        # Set timeout
        oldtime = time.time()
        if timeout=="" or timeout=="undef":
            timeout = float(link["timeout"])
        else:
            timeout = float(timeout)
        # Read until pattern
        buffer = "\x00" * len(pattern)
        while buffer != pattern and timeout >= 0:
            buffer = buffer[1:]
            res = '\x00'
            while res == '\x00' and timeout >= 0:
                try:
                    retcode,res = self.ll_receive(self.ll_links[link["descr"]],link,1,timeout)
                except Exception as e:
                    return 0,"Pattern not found or %s"%(e)
                if retcode == 0:
                    return 0,res
                timeout -= time.time() - oldtime
                oldtime = time.time()
            buffer += res
            if (DEBUG):
                if ord(res)<32: sys.stdout.write(" "+str(ord(res))+" ")
                else: sys.stdout.write(res)
        if timeout < 0:
            return 0,"Pattern not found"
        return 1,"Pattern found"

