/*
Copyright 2012-2015 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <dlfcn.h>
#include <time.h>
#include <ctype.h>
#include <semaphore.h>
#include <pyrame.h>

#include <cmdmod_env.h>
#include <cmdmod_wrapper.h>
#include <cmdmod_dispatcher.h>
#include <cmdmod_task.h>

#ifndef CMDMOD_H
#define CMDMOD_H

#define NEXT_HOST   1
#define NEXT_PORT   2
#define NEXT_FILE   3
#define NEXT_IGNORE 5
#define NEXT_LPORT  6
#define NEXT_LANG   7
#define NEXT_NAME   8

#define SUBMOD_TYPE_NONE 0
#define SUBMOD_TYPE_SCRIPT 1
#define SUBMOD_TYPE_INT 2
#define SUBMOD_TYPE_HOST 3
#define SUBMOD_TYPE_CMOD 4
#define SUBMOD_TYPE_ERROR 5

#define FUNC_ERROR_NOTINCMOD 0
#define FUNC_ERROR_NOCMOD 1

#define UCMD_DELAY 2000000

//pyrame result retcode
#define ERROR 0
#define OK 1
#define WAKEUP 2
#define KEEPOPEN 3

struct bg_ws {
  int index;
  void *params;
};

struct cmdmod_ws;
struct function;

struct func_db {
  struct function *first;
};

struct conn {
  char *host;
  int port;
  int socket;
  pthread_mutex_t *use_mutex;
};

struct conn_db {
  struct conn *conn;
  struct conn_db *next;
};

//the structure describe the functions of the module as stated in xml file
struct function {
  //name of the function
  char *name;
  //type could be script(executed locally) or host(executed on another module)
  unsigned char type; //SUBMOD_TYPE_*
  //for script functions
  char *api;
  //for host function
  char *host;
  int port;
  char *portstring;
  struct conn *conn;
  //for cmod function
  char *id;
  //in case of error
  char *error_msg;
  //for internal functions
  struct cmd_result *(*fp)(struct cmdmod_ws *ws,struct cmd *command);
  //chaining
  struct function *next;
};

struct cmdmod_ws {
  //language specific workspace
  void *lang_spec_ws;
  //language specific functions
  void (*deinit_func)(void *);
  struct cmd_result *(*execcmd_func)(struct cmd *command,struct function *func,void *ws);
  //global structs
  struct func_db *hfdb; //host func db
  struct func_db *sfdb; //script func db
  struct func_db *cfdb; //cmod func db
  struct conn_db *cdb; //conn db
  struct function *error_func;
  struct wrapenv *we;
  //netserv
  struct netserv_client *clients;
  //parsing variables
  struct function *current_function;
  int next_data;
  //cmod function cache
  char local;
  int local_next_id;
  char *cmod_ip;
  unsigned short cmod_port;
  //current socket
  int current_socket;
  int current_index;
  //output redirecting
  pthread_t *t_output; 
  int pipe_out[2];
  int old_out;
  //init args
  char **init_args;
  int init_nb_args;
  //environment
  struct cenv *env;
  //execution mutex
  pthread_mutex_t *exec_mutex;
  //dispatcher
  struct dispatch_ws *dws;
  //debug
  char debug;
};

void debug(const char *format, ...);
void output(struct cmdmod_ws *ws,const char *format, ...);
struct cmd_result * treat_cmd(struct cmd *command);
char *get_caller_ip(void);

//connections database
struct conn *get_conn(struct cmdmod_ws *ws,char *host,int port);
void close_all_conns(struct cmdmod_ws *ws);

//function databases
struct func_db *init_func_db(void);
void debug_db(struct func_db *db);
struct function *search_func_byname(struct func_db *db,char *name);
void add_func(struct func_db *db,struct function *new_func);

//specific function databases
void print_func(struct function *func);
struct function *search_cmod_function_in_db(struct cmdmod_ws *ws,char *name,char *id);
struct function *ask_cmod(struct cmdmod_ws *ws,char *funcname,char *module,char *id);
char *dump_script_func_db(struct cmdmod_ws *ws);

//internal function database
void init_ifdb(struct func_db *db,char *name);


//exit properly
void cexit(int nb);

//cmdmod_treat
struct cmd_result * treat_intcmd(struct cmdmod_ws *ws,struct cmd *command);
struct cmd_result * treat_extcmd(struct cmdmod_ws *ws,struct cmd *command);
struct cmd_result * treat_ext_function(struct cmdmod_ws *ws,struct cmd *command,struct function* func);
struct function *search_ext_function(struct cmdmod_ws *ws,struct cmd *command);
struct function *search_script_function(struct cmdmod_ws *ws,struct cmd *command);
struct function *create_hfdb_function(struct cmdmod_ws *ws,char *name,char *module,char *ip);

//internal functions
struct cmd_result *treat_exit(struct cmdmod_ws *ws,struct cmd *command);
struct cmd_result * treat_register(struct cmdmod_ws *ws,struct cmd *command);
struct cmd_result * treat_getfuncapi(struct cmdmod_ws *ws,struct cmd *command);
struct cmd_result * treat_getapi(struct cmdmod_ws *ws,struct cmd *command);
struct cmd_result * treat_printenv(struct cmdmod_ws *ws,struct cmd *command);

//language specific functions 

void *init_dummy(struct cmdmod_ws *cws);
void *init_python(struct cmdmod_ws *cws);
void *init_lua(struct cmdmod_ws *cws);
void *init_c(struct cmdmod_ws *cws);
void *init_sh(struct cmdmod_ws *cws);
void *init_R(struct cmdmod_ws *cws);
void deinit_dummy(void *ws);
void deinit_python(void *ws);
void deinit_lua(void *ws);
void deinit_c(void *ws);
void deinit_sh(void *ws);
void deinit_R(void *ws);

struct cmd_result *exec_command_dummy(struct cmd *command,struct function *func,void *ws);
struct cmd_result *exec_command_python(struct cmd *command,struct function *func,void *ws);
struct cmd_result *exec_command_lua(struct cmd *command,struct function *func,void *ws);
struct cmd_result *exec_command_c(struct cmd *command,struct function *func,void *ws);
struct cmd_result *exec_command_sh(struct cmd *command,struct function *func,void *ws);
struct cmd_result *exec_command_R(struct cmd *command,struct function *func,void *ws);

//submod functions
struct cmdmod_ws *cmdmod_getenv(void);
struct cmd_result *submod_execcmd(char * funcname,...);
int submod_getport(char *port_name);
char *submod_getip(char *func_name,int id);
void submod_exit(int exit_code);
void submod_setmyname(char *newname);

//thread
struct bg_ws *submod_bgcontext(void);
void submod_sendres(struct bg_ws *context,struct cmd_result *cmdres);

//env
int submod_setvar(const char *ns,const char *name,char *value,char type);
char *submod_getvar(const char *ns,const char *name);
char submod_gettype(const char *ns,const char *name);
int submod_delvar(const char *ns,const char *name);
int submod_newns(const char *ns);
int submod_checkns(const char *ns);
int submod_delns(const char *ns);
char *submod_dumpns(const char *ns);
char *submod_listns(void);

//task
int submod_wakemeup(int time,char *function,char *param);

//dispatcher
int submod_init_dispatcher(int nb_blocks,char *name,char *schema_str,char *port);
int submod_flush_dispatcher(void);
int submod_new_block(int block_id,struct block **result);
int submod_new_event(struct block *bl,struct event **result);
void submod_add_block(struct block *bl,int block_id);
int submod_set_field_block(struct block *bl,struct evtstructelem *ese,char *format,...);
int submod_set_field_event(struct event *ev,struct evtstructelem *ese,char *format,...);
int submod_finalize_block(struct block *bl);
int submod_set_chans(char *schema_str);

//helper emulated function
void submod_setres(int retcode, char *format,...);
struct cmd_result *submod_setmyport(unsigned short myport);
struct cmd_result *submod_getmyport(void);
struct cmd_result *submod_getmyfile(void);

#endif
