/*
Copyright 2012-2014 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/


#include "cmdmod_lua.h"

static int submod_exec_lua(lua_State *L) {
  int i;
  struct cmd *command=malloc(sizeof(struct cmd));
  struct cmd_result *cres;

  int argc=lua_gettop(L);
  //printf("submod_exec called with %d arguments\n",argc);

  //extract the command name
  command->name=strdup(lua_tostring(L,1));

  //extract the command params
  command->nb_params=argc-1;
  command->params=malloc(command->nb_params*sizeof(char *)); 
  for (i=0;i<argc-1;i++) {
    command->params[i]=strdup(lua_tostring(L,i+2));
  }

  //treat the command
  cres=treat_cmd(command);

  //push the result in the lua stack
  lua_pushnumber(L,cres->retcode);
  lua_pushstring(L,cres->str);

  //free the structs
  free_cmd(command);
  free_cmd_result(cres);

  //return to lua
  return 2;
}

static int submod_callwrap_lua(lua_State *L) {

  char *func_name;
  char **wargs;
  int i;
  struct submod_wrapper *wrapper;
  struct cmd_result *pres;
  struct cmdmod_ws *ws=cmdmod_getenv();

  func_name=(char *)luaL_checklstring(L,1,NULL);


  wrapper=lookup_wrapper(ws->we,func_name);
  if (wrapper==NULL) {
    lua_pushnumber(L,0);
    lua_pushstring(L,"unknown function");
    return 2;
  }

  wargs=malloc(wrapper->nb_args*sizeof(char *));
  for(i=0;i<wrapper->nb_args;i++) {
    wargs[i]=(char *)luaL_checklstring(L,i+2,NULL);
  }
  lua_pop(L,1+wrapper->nb_args);

  //call the wrapper
  pres=wrapper->wrapper(wargs);
  free(wargs);

  lua_pushnumber(L,pres->retcode);
  lua_pushstring(L,pres->str);
  free_cmd_result(pres);
  return 2;
}


struct cmd_result *exec_command_lua(struct cmd *command,struct function *func,void *ws) {
  int i;
  int res;
  struct lua_ws *lws=(struct lua_ws *)ws;
  int retcode;
  char *retstr;

  //find the function in the lua environment
  lua_getglobal(lws->L,func->name);
  
  //pass the arguments
  for (i=0;i<command->nb_params;i++)
    lua_pushstring(lws->L,command->params[i]);
  
  if (lua_isfunction(lws->L,lua_gettop(lws->L)-command->nb_params)) {
    //call the lua function
    res=lua_pcall(lws->L,command->nb_params,2,0);
    if (res==0) {
      retcode=luaL_checknumber(lws->L,1);
      retstr=(char *)luaL_checklstring(lws->L,2,NULL);
      lua_pop(lws->L,2);
      return raw_build_result(retcode,retstr);
    } else {
      return build_result(0,"Error in running  : %s\n",lua_tostring(lws->L,-1));
    }
  } else {
    return build_result(0,"Error : unknown function %s\n",func->name);
  }
}


void exec_file_lua(struct lua_ws *lws,char *filename) {
  FILE *f;
  
  //sanity checks on the code file
  if (filename==NULL) {
    printf("Error : undefined lua file name...exiting\n");
    cexit(1);
  }
  f=fopen(filename,"r");
  if (f==NULL) {
    printf("Error : unable to open file %s...exiting\n",filename);
    fclose(f);
    cexit(1);
  }
  fclose(f);
  
  //execute the code file
  if (luaL_loadfile(lws->L,filename)) {
    printf("Error : syntax error in %s...exiting\n",filename);
    cexit(1);
  }
  if (lua_pcall(lws->L,0,LUA_MULTRET,0)) {
    printf("Error : cant execute file %s : error %s...exiting\n",filename,lua_tostring(lws->L,-1));
    cexit(1);
  }
}


void *init_lua(struct cmdmod_ws* cws) {
  int res;
  int i;
  //create the language specific workspace
  struct lua_ws *lws=malloc(sizeof(struct lua_ws));
  memset(lws,0,sizeof(struct lua_ws));
  lws->L=luaL_newstate();

  //load libs and submod funcs
  luaL_openlibs(lws->L);
  lua_pushcfunction(lws->L,submod_exec_lua);
  lua_setglobal(lws->L,"submod_execcmd");
  lua_pushcfunction(lws->L,submod_callwrap_lua);
  lua_setglobal(lws->L,"submod_callwrap");

  //execute helper
  exec_file_lua(lws,"/opt/pyrame/cmdmod_helper.lua");
  
  //execute module code
  exec_file_lua(lws,submod_getvar("system","code_file_name"));

  //execute init function
  lua_getglobal(lws->L,"init");
  if (lua_isfunction(lws->L,lua_gettop(lws->L))) {
    for(i=0;i<cws->init_nb_args;i++) {
      lua_pushstring(lws->L,cws->init_args[i]);
    }
    printf("pushing %d arguments\n",cws->init_nb_args);
    res=lua_pcall(lws->L,cws->init_nb_args,0,0);
    if (res!=0) {
      printf("Error in running init function : %s\n",lua_tostring(lws->L,-1));
      cexit(1);
    }
  }

  return (void *)lws;
}

void deinit_lua(void *ws) {
  struct lua_ws *lws=(struct lua_ws *)ws;
  int res;
  //execute deinit function
  lua_getglobal(lws->L,"deinit");
  if (lua_isfunction(lws->L,lua_gettop(lws->L))) {
    res=lua_pcall(lws->L,0,0,0);
    if (res!=0) {
      printf("Error in running deinit function : %s\n",lua_tostring(lws->L,-1));
      return;
    }
  }
}
