/*
 * Copyright 2012-2015 Frédéric Magniette, Miguel Rubio-Roy
 * This file is part of Pyrame.
 *
 * Pyrame is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pyrame is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
 * */

#include "cmdmod_python.h"

static PyObject* submod_execcmd_python(PyObject *self,PyObject *args) {
  PyObject* result;
  
  int i;
  struct cmd *command=malloc(sizeof(struct cmd));
  struct cmd_result *cres;
  PyObject*pyres;

  pyres=PyTuple_GetItem(args,0);
  if (pyres==NULL) {
    printf("Error : cant get python function name\n");
    free(command);
    return NULL;
  }
  command->name=PyString_AsString(pyres);
  if (command->name==NULL) {
    printf("Error : cant get python function name\n");
    free(command);
    return NULL;
  }
  command->nb_params=PyTuple_Size(args)-1;
  command->params=malloc(command->nb_params*sizeof(char *));
  for (i=0;i<command->nb_params;i++) {
    pyres=PyTuple_GetItem(args,i+1);
    if (pyres==NULL) {
      printf("Error : cant get python parameter number %d\n",i);
      free(command->params);
      free(command);
      return NULL;
    }
    command->params[i]=PyString_AsString(pyres);
    if (command->params[i]==NULL) {
      printf("Error : cant get python parameter number %d\n",i);
      free(command->params);
      free(command);
      return NULL;
    }
  }

  //treat the command
  cres=treat_cmd(command);
  result=Py_BuildValue("is",cres->retcode,cres->str);

  //free the command : CAUTION : dont free parameters or function name
  free(command->params);
  free(command);
  free_cmd_result(cres);
  return result;
}

static PyObject* submod_exit_python(PyObject *self,PyObject *args) {
  int sig;
  if(!PyArg_ParseTuple(args,"i",&sig)) {
    printf("cant parse signal param\n");
    return NULL;
  } 
  submod_exit(sig);
  return NULL;
}

static PyObject* submod_callwrap_python(PyObject *self,PyObject *args) {
  PyObject* result;
  char *func_name;
  PyObject*pyres;
  int i;
  struct submod_wrapper *wrapper;
  char **wargs;
  struct cmd_result *pres;
  char msg[2048];
  struct cmdmod_ws *ws=cmdmod_getenv();

  pyres=PyTuple_GetItem(args,0);
  if (pyres==NULL) {
    printf("Error : cant get submod function name\n");
    return Py_BuildValue("is",0,"Cant get function name");
  }
  func_name=PyString_AsString(pyres);
  
  //lookup for wrapper
  wrapper=lookup_wrapper(ws->we,func_name);
  if (wrapper==NULL) {
    return Py_BuildValue("is",0,"unknown function");
  }
  
  //build the char **args
  wargs=malloc(wrapper->nb_args*sizeof(char *));
  for(i=0;i<wrapper->nb_args;i++) {
    pyres=PyTuple_GetItem(args,i+1);
    if (pyres==NULL) {
      printf("Error : cant get argument %d\n",i+1);
      sprintf(msg,"Error : cant get argument %d",i+1);
      return Py_BuildValue("is",0,msg);
    }
    wargs[i]=PyString_AsString(pyres);
    if (wargs[i]==NULL) {
      sprintf(msg,"Error : argument %d is not a string",i+1);
      return Py_BuildValue("is",0,msg);
    }
  }

  //call the wrapper
  pres=wrapper->wrapper(wargs);
  free(wargs);

  result=Py_BuildValue("is",pres->retcode,pres->str);
  free_cmd_result(pres);
  return result;
}

static PyObject* submod_getport_python(PyObject *self,PyObject *args) {
  PyObject* result;
  int port;
  char *req;

  if(!PyArg_ParseTuple(args,"s",&req)) {
    return NULL;
  }
  port=submod_getport(req);
  result=Py_BuildValue("i",port);
  return result;
}

char *build_txtargs_py(int argc,char **argv) {
  int i;
  struct incr_string *ires=new_is();
  fsuffix_is(ires,"(");
  for(i=0;i<argc;i++) {
    if (i==0) {
      fsuffix_is(ires,"r\"%s\"",argv[i]);
    } else {
      fsuffix_is(ires,",r\"%s\"",argv[i]);
    }
  }
  fsuffix_is(ires,")");
  return reduce_is(ires);
}

void exec_file_python(struct python_ws *pws,char *filename) {
  FILE *f;
  PyObject *pyres;
  
  //execute the file
  //printf("loading the file %s...",filename);
  if (filename==NULL) {
    printf("Error : empty filename...exiting\n");
    cexit(1);
  }
  f=fopen(filename,"r");
  if (f==NULL) {
    printf("Error : unable to open file %s...exiting\n",filename);
    cexit(1);
  }
  //execute the code file
  pyres=PyRun_File(f,filename,Py_file_input,pws->dict,pws->dict);
  if (pyres==NULL) {
    fclose(f);
    printf("Error : syntax error in %s...exiting\n",filename);
    PyErr_Print();
    cexit(1);
  }
  fclose(f);
  //printf("done\n");
}

void *init_python(struct cmdmod_ws *cws) {
  /* this function is called by cmdmod if language is NULL or python
     cws is the cmdmod workspace to save in python environment for submod.exec usage.
  */
  char *cmd;
  char *params;
  PyObject *pyres;

  //create the language specific workspace
  struct python_ws *pws=malloc(sizeof(struct python_ws));
  memset(pws,0,sizeof(struct python_ws));
  pws->cws=cws;
  pws->id=1234;

  //initialize the language virtual machine
  Py_Initialize();

  //creating dictionaries
  pws->main=PyImport_AddModule("__main__");
  pws->dict=PyModule_GetDict(pws->main);

  //register the submod functions
  Py_InitModule("submod",submodmethods);
  pyres=PyRun_String("import submod",Py_single_input,pws->dict,pws->dict);
  if (pyres==NULL) {
    printf("cant import submod\n");
    PyErr_Print();
    cexit(11);
  }

  //execute helper
  exec_file_python(pws,"/opt/pyrame/cmdmod_helper.py");
  
  //execute module code
  exec_file_python(pws,submod_getvar("system","code_file_name"));
  
  //execute the init function
  params=build_txtargs_py(cws->init_nb_args,cws->init_args);
  cmd=(char *)malloc(strlen(params)+100);
  sprintf(cmd,"if \"init\" in dir() and hasattr(init,\"__call__\"):\n  init%s",params);
  free(params);
  pyres=PyRun_String(cmd,Py_single_input,pws->dict,pws->dict);
  if (pyres==NULL) {
    printf("error during init\n");
    PyErr_Print();
    cexit(12);
  }
  free(cmd);
    
  pws->_save=PyEval_SaveThread();
  return (void *)pws;
}

void deinit_python(void *ws) {
  struct python_ws *pws=(struct python_ws *)ws;
  char *cmd=(char *)malloc(1024);
  sprintf(cmd,"if \"deinit\" in dir() and hasattr(deinit,\"__call__\"):\n  deinit()");
  PyEval_RestoreThread(pws->_save);
  PyRun_SimpleString(cmd);
  pws->_save=PyEval_SaveThread();
  free(cmd);
}

PyObject *build_args_py(int argc,char **argv) {
  int i;
  PyObject *params,*pystr;
  params=PyTuple_New(argc);
  for (i=0;i<argc;i++) {
    pystr=PyString_FromString(argv[i]);
    if (PyTuple_SetItem(params,i,pystr)) {
      return NULL;
    }
  }
  return params;
}

struct cmd_result *exec_command_python(struct cmd *command,struct function *func,void *ws) {
  /* This function is called any time a pyrame command is received by cmdmod.
     command contains the name of the pyrame command and the parameters
     func contains the description of the function including its python func name.
  */
  
  PyObject *params;
  PyObject *pyres;
  char *retstr;
  char *retcode_str;
  int retcode=1;
  struct python_ws *pws=(struct python_ws *)ws;
  PyObject *pyfunc;
  PyObject *etype,*etraceback,*evalue;
  char *error_msg;
  PyTracebackObject* traceback;
  int lineno=0;
  char *filename=NULL;
  char *funcname=NULL;
  struct cmd_result *pres;

  //free the setres env
  env_setvar(pws->cws->env,"system","retcode","0",ENV_INT,0);
  env_setvar(pws->cws->env,"system","retstr","No answer from script",ENV_STR,0);
  
  PyEval_RestoreThread(pws->_save);
  pyfunc=PyObject_GetAttrString(pws->main,func->name);
  if (pyfunc==NULL || !PyCallable_Check(pyfunc)) {
    return build_result(0,"function %s does not exist\n",func->name);
  }
  params=build_args_py(command->nb_params,command->params);
  if (params==NULL) {
    return raw_build_result(0,"Unable to build python params tuple");
  }
  pyres=PyObject_CallObject(pyfunc,params);
  Py_DECREF(params); 

  if (pyres==NULL) {
    //PyErr_Print();
    PyErr_Fetch(&etype,&evalue,&etraceback);
    if (evalue==NULL) {
      pws->_save=PyEval_SaveThread();
      Py_XDECREF(etype);
      return raw_build_result(0,"NULL exception raised in Python code");
    }
    traceback=(PyTracebackObject*)etraceback;
    printf("Traceback (most recent call last):\n");
    while (traceback!=NULL) {
      lineno=traceback->tb_lineno;
      filename=PyString_AsString(traceback->tb_frame->f_code->co_filename);
      funcname=PyString_AsString(traceback->tb_frame->f_code->co_name);
      printf("  File %s, line %d, in %s\n",filename,lineno,funcname);
      traceback=traceback->tb_next;
    }
    error_msg=PyString_AsString(PyObject_Str(evalue));
    Py_DECREF(etype);
    //Py_DECREF(etraceback);
    Py_DECREF(evalue);
    printf("ValueError: %s\n",error_msg);
    pws->_save=PyEval_SaveThread();
    return build_result(0,"Exception raised in Python code : %s at line %d in %s in %s",error_msg,lineno,funcname,filename);
  }

  if (pyres==Py_None) {
    retcode_str=env_getvar(pws->cws->env,"system","retcode");
    retstr=env_getvar(pws->cws->env,"system","retstr");
    return raw_build_result(atoi(retcode_str),retstr);
  }

  if (!PyTuple_Check(pyres)) {
    return raw_build_result(0,"Script not returning tuple");
  }

  if (PyTuple_Size(pyres)!=2) {
    return build_result(0,"Script giving bad result (%d entries)",PyTuple_Size(pyres));
  }

  if(!PyArg_ParseTuple(pyres,"is",&retcode,&retstr)) {
    return raw_build_result(0,"Unable to parse Python result");
  }

  //printf("retstr:%s\n",retstr);
  pres=raw_build_result(retcode,retstr);
  Py_DECREF(pyres);
  pws->_save=PyEval_SaveThread();
  return pres;
}

