/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include <cmd_converter.h>

void conv_usage() {
  printf("parameters : ds_name port uri convert_plugin\n");
  printf("uri can be one of the following : \n");
  printf("shmem://shmemid\n");
  printf("tcp://host:port\n");
  printf("file://path/filename\n");
  printf("raw2root://path/filename\n");
  submod_exit(1);
}

//the thread that converts
void *converter(void *args) {
  struct conv_ws *ws=(struct conv_ws *)args;
  ws->convert(ws);
  return NULL;
}


void *init(int argc,char **argv) {
  char *protocol=NULL;
  int i;
  char *uri;
  char *newname;
  int res;
  
  //init the workspace
  struct conv_ws * ws=malloc(sizeof(struct conv_ws));
  bzero(ws,sizeof(struct conv_ws));

  //extract parameters from command line
  if (argc<4) {
    printf("4 args mandatory, %d provided\n",argc);
    conv_usage();
  }

  ws->ds_name=strdup(argv[0]);
  ws->cmd_port=strdup(argv[1]);
  uri=argv[2];
  ws->plugin_name=strdup(argv[3]);
  ws->argc=argc-4;
  ws->argv=argv+4;
  //printf("argc=%d ws->argc=%d\n",argc,ws->argc);

  //parse uri
  for(i=0;uri[i]!=0;i++) {
    if (uri[i]==':') {
      protocol=strndup(uri,i);
      i+=3;
      break;
    }
  }
  uri+=i;

  ws->itype=UNKNOWN;
  if (protocol==NULL) {
    printf("syntax error in uri : %s\n",uri);
    submod_exit(1);
  }

  if (!strcmp(protocol,"shmem")) {
    ws->itype=SHMEM;
    ws->output_mode=ONLINE;
    sscanf(uri,"%d",&(ws->shmemid));
    printf("read shmemid=%d\n",ws->shmemid);
  }

  if (!strcmp(protocol,"tcp")) {
    ws->itype=TCP;
    ws->output_mode=ONLINE;
    for(i=0;uri[i]!=0;i++) {
      if (uri[i]==':') {
        ws->host=strndup(uri,i);
        ws->iport=atoi(uri+i+1);
        break;
      }
    }
    if (ws->host==NULL) {
      printf("syntax error (no host found)\n");
      conv_usage();
    }
  }

  if (!strcmp(protocol,"file")) {
    ws->itype=OFILE;
    ws->filename=strdup(uri);
    ws->output_mode=ONLINE;
  }

  if (!strcmp(protocol,"raw2root")) {
    ws->itype=OFILE;
    ws->filename=strdup(uri);
    ws->output_mode=OFFLINE;
  }
  
  if (protocol==NULL) {
    printf("unable to extract protocol from uri\n");
    printf("\n\n");
    conv_usage();
  }
  if (ws->itype==UNKNOWN) {
    printf("unknown protocol %s\n",protocol);
    printf("\n\n");
    conv_usage();
  }

  switch(ws->itype) {
  case SHMEM:
    //open the shared buffer
    printf("open and ready shbuf %d\n",ws->shmemid);
    ws->shbuf=open_shbuffer(ws->shmemid,NO_CREATE,0);
    ready_shbuffer(ws->shbuf);
    break;
  case TCP:
    ws->data_socket=open_socket(ws->host,ws->iport);
    if (ws->data_socket==-1) {
      printf("unable to open data connection...exiting\n");
      submod_exit(1);
    }
    break;
  case OFILE:
    ws->finput=fopen(ws->filename,"r");
    if (ws->finput==NULL) {
      printf("unable to open file %s...exiting\n",ws->filename);
      submod_exit(1);
    }
    break;
  default:
    printf("unknown protocol %s\n",protocol);
    conv_usage();
  }

  //set the name
  newname=malloc(128+strlen(ws->ds_name));
  sprintf(newname,"converter_%s",ws->ds_name);
  submod_setmyname(newname);

  //load the convert plugin
  ws->lib=dlopen(ws->plugin_name,RTLD_NOW|RTLD_LOCAL);
  if (ws->lib==NULL) {
    printf("plugin dlopen error : %s...exiting\n",dlerror()); 
    submod_exit(1);
  }
  ws->convert=dlsym(ws->lib,"convert");
  if (ws->convert==NULL) {
    perror("dlsym");
    printf("cant find symbol convert...exiting\n");
    submod_exit(1);
  }
  ws->init=dlsym(ws->lib,"init_plugin");
  if (ws->init==NULL) {
    printf("cant find symbol init...exiting\n");
    submod_exit(1);
  }
  ws->reinit=dlsym(ws->lib,"reinit");
  if (ws->reinit==NULL) {
    printf("cant find symbol reinit...exiting\n");
    submod_exit(1);
  }
  
  
  //init the stats system
  printf("init stats\n");
  ws->stats=init_stats(submod_getip("getvar@varmod",0),1000000);

  //fill the scheme
  
  ws->init(ws);
  if (ws->schema_str==NULL) {
    printf("warning : converter : schema_str is null ...exiting\n");
    submod_exit(1);
  }
  if (ws->schema==NULL) {
    ws->schema=saf_parse_chans(ws->schema_str);
  }

  res=submod_init_dispatcher(BLOCK_NB,newname,ws->schema_str,ws->cmd_port);
  if (res<0) {
    printf("error : %s\n",pyrame_errors[-res]);
    submod_exit(1);
  }

  //set the command port
  submod_setmyport(res);
  
  return ws; 
}


void deinit(void *workspace) {
  struct conv_ws *ws=(struct conv_ws *)workspace;
  printf("converter deinit\n");
  deinit_stats(ws->stats);
  if (ws->t_converter!=NULL)
    pthread_cancel(*ws->t_converter);
  printf("converter exiting\n");
}


//pyrapi:flush_converter:
struct cmd_result *flush_converter(struct cmd *command,void *workspace) {
  struct conv_ws *ws=(struct conv_ws *)workspace;
  printf("flush converter\n");
  submod_flush_dispatcher();
  flush_ribbon(ws);
  return build_result(1,"ok");
}

//pyrapi:start_acq_converter:
struct cmd_result * start_acq_converter(struct cmd *command,void *workspace) {
  struct conv_ws *ws=(struct conv_ws *)workspace;
  
  debug("start_acq converter\n");

  //flush internal circular buffer
  submod_flush_dispatcher();

  //flush data ribbon
  flush_ribbon(ws);

  //reinit the decoding plugin
  ws->reinit(ws);
  
  //launching converting thread
  ws->ready=0;
  printf("launching converting thread\n");
  ws->t_converter=(pthread_t*)malloc(sizeof(pthread_t));
  if (pthread_create(ws->t_converter,NULL,converter,(void *)ws)<0) {
    perror("pthread_create");
    exit(1);
  }
  pthread_detach(*ws->t_converter);
  
  while(!ws->ready) {
    usleep(10000);
  }

  if (ws->itype==SHMEM) {
    ready_shbuffer(ws->shbuf);
  }
  
  return build_result(1,"ok");
}

//pyrapi:stop_acq_converter:
struct cmd_result *stop_acq_converter(struct cmd *command,void *workspace) {
  struct conv_ws *ws=(struct conv_ws *)workspace;

  //kill the converting thread
  printf("stopping converting thread\n");
  if (ws->t_converter!=NULL)
    pthread_cancel(*ws->t_converter);
  ws->t_converter=NULL;
  
  return build_result(1,"ok");
}

//pyrapi:reset_stats_converter:
struct cmd_result *reset_stats_converter(struct cmd *command,void *workspace) {
  struct conv_ws *ws=(struct conv_ws *)workspace;
  zero_stats(ws->stats);
  return build_result(1,"ok");
}

