/*
 * Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
 * This file is part of Pyrame.
 *
 * Pyrame is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pyrame is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
 * */

#include <cmd_converter.h>

#define max(a,b) a<b?b:a

void init_ribbon(struct conv_ws *ws,int size) {
  ws->buffer=(char *)malloc(size);
  ws->buffer_size=size;
  ws->data_start=ws->buffer;
  ws->data_size=0;
}

void flush_ribbon(struct conv_ws *ws) {
  ws->data_start=ws->buffer;
  ws->data_size=0;
}

void set_debug(struct conv_ws *ws) {
  ws->debug=1;
}

void print_char(char c) {
  /*if (c>=32 && c<=126) {
    printf("\[%c",c);
    return;
    }*/
  switch(c) {
  case '\n':
    printf("[\\n]");
    break;
  default:
    printf("%02x",c&0xff);
  }
}

void print_string(char *str,int size) {
  int i;
  for(i=0;i<size;i++) {
    print_char(str[i]);
  }
}

void print_ribbon(struct conv_ws *ws) {
  int max_size,i;
  int limit=12;
  printf("ribbon size=%d data=",ws->data_size);
  if (ws->data_size<limit) {
    max_size=ws->data_size;
  } else {
    max_size=limit;
  }
  for(i=0;i<max_size;i++) {
    print_char(ws->data_start[i]);
  }
  if (max_size==limit) {
    printf("...\n");
  } else {
    printf("\n");
  }
}

int ensure_data(struct conv_ws *ws,int size) {
  int rsize;
  while (ws->data_size<size) {
    if (ws->debug) {
      printf("ensure_data : acquiring data because data_size=%d\n",ws->data_size);
    }
    rsize=request_new_data(ws,((char *)(ws->data_start+ws->data_size)));
    if (rsize==EBROK) {
      return EBROK;
    } else {
      ws->data_size+=rsize;
      if (ws->debug) {
        printf("acquired %d bytes total=%d\n",rsize,ws->data_size);
        print_ribbon(ws);
      }
    }
  }
  return 0;
}

int forget_before_current(struct conv_ws *ws) {
  char *cpy;
  if (ws->debug) {
    printf("forget before current\n");
  }
  if (ws->data_size>0) {
    cpy=(char *)malloc(ws->data_size);
    memcpy(cpy,ws->data_start,ws->data_size);
    memcpy(ws->buffer,cpy,ws->data_size);
    free(cpy);
  }
  ws->data_start=ws->buffer;
  return 0;
}

int seek_forward(struct conv_ws *ws,int size) {
  //ensure that at least "size" data are available
  if (ensure_data(ws,size)==EBROK)
    return EBROK;

  if (ws->debug) {
    printf("moving forward of %d\n",size);
  }
  
  //forward on the ribbon
  ws->data_start+=size;
  ws->data_size-=size;
  if (ws->debug) {
    printf("result of forward : \n");
    print_ribbon(ws);
  }
  return 0;
}

int seek_backward(struct conv_ws *ws,int size) {
  //backward on the ribbon : CAUTION : no verification about available previous data

  if (ws->debug) {
    printf("moving backward of %d\n",size);
  }
  
  ws->data_start-=size;
  ws->data_size+=size;
  if (ws->debug) {
    printf("result of backward : \n");
    print_ribbon(ws);
  }
  return 0;
}

int read_nb(struct conv_ws *ws,char **result,int size) {
  //ensure that at least "size" data are available
  if (ensure_data(ws,size)==EBROK)
    return EBROK;

  //copy the result
  *result=(char *)malloc(size+1);
  memcpy(*result,ws->data_start,size);
  (*result)[size]=0;
  //forward on the ribbon
  ws->data_start+=size;
  ws->data_size-=size;
  return size;
}


int forward_until(struct conv_ws *ws,char term) {
  if (ws->debug) {
    printf("forward until ");
    print_char(term);
    printf("\n");
  }
  do {
    if (ensure_data(ws,1)==EBROK) {
      return EBROK;
    }
    
    if (ws->data_start[0]==term) {
      if (ws->debug) {
        printf("found term ");
        print_char(term);
        printf("\n");
        print_ribbon(ws);
      }
      return 0;
    }
    
    //forward on the ribbon
    ws->data_start+=1;
    ws->data_size-=1;
  } while(1);
  //never reached
  return 0;
}

int cur_char(struct conv_ws *ws,char *c) {
  if (ensure_data(ws,1)==EBROK)
      return EBROK;
  
  *c=ws->data_start[0];
  if (ws->debug) {
    printf("cur_char ");
    print_char(*c);
    printf("\n");
  }
  return 0;
}
  

int read_until(struct conv_ws *ws,char **result,char term1,char avoid) {
  int idx=0;
  
  if (ws->debug) {
    printf("read_until ");
    print_char(term1);
    printf(" avoid ");
    print_char(avoid);
    printf(" on \n");
    print_ribbon(ws);
  }
  
  do {

    if (ensure_data(ws,idx+1)==EBROK)
      return EBROK;

    if (ws->data_start[idx]==avoid) {
      *result=(char *)malloc(idx+2);
      memcpy(*result,ws->data_start,idx+1);
      (*result)[idx+1]=0;
      return EAVOID;
    }
    
    if (ws->data_start[idx]==term1) {
      *result=(char *)malloc(idx+1);
      memcpy(*result,ws->data_start,idx);
      (*result)[idx]=0;
      ws->data_start+=idx;
      ws->data_size-=idx;
      return idx;
    }

    idx++;
     
  } while(1);
  //never reached
  return 0;
}

int read_until_str(struct conv_ws *ws,char **result,char *term_str,char *avoid_str) {
  int min_size;
  int term_size;
  int avoid_size;
  int res_size;

  term_size=strlen(term_str);
  if (avoid_str)
    avoid_size=strlen(avoid_str);
  else
    avoid_size=0;
  min_size=max(term_size,avoid_size);

  if (ws->debug) {
    printf("read until string ");
    print_string(term_str,term_size);
    if (avoid_str) {
      printf(" avoiding ");
      print_string(avoid_str,avoid_size);
    }
    printf("\n");
  }

  
  int idx=0;
  do {

    if (ensure_data(ws,idx+min_size)==EBROK)
      return EBROK;
    
    if (avoid_str!=NULL) {
      if (!strncmp(ws->data_start+idx,avoid_str,strlen(avoid_str))) {
        if (ws->debug)
          printf("avoid found at idx=%d\n",idx);
	return EAVOID;
      }
    }
    
    if (!strncmp(ws->data_start+idx,term_str,strlen(term_str))) {
      if (ws->debug)
        printf("found at idx=%d\n",idx);
      res_size=idx+min_size;
      *result=(char *)malloc(res_size+1);
      memcpy(*result,ws->data_start,res_size);
      (*result)[res_size]=0;
      ws->data_start+=res_size;
      ws->data_size-=res_size;
      return res_size;
    }

    idx++;
     
  } while(1);
  return 0;
}

int forward_until_str(struct conv_ws *ws,char *term_str,char *avoid_str) {
  int min_size=strlen(term_str);
  int avoid_size;

  if (avoid_str!=NULL)
     avoid_size=strlen(avoid_str);
  else
    avoid_size=0;

  if (ws->debug) {
    printf("forward until string ");
    print_string(term_str,min_size);
    if (avoid_str) {
      printf(" avoiding string ");
      print_string(avoid_str,avoid_size);
    }
    printf("\n");
  }

  int idx=0;
  do {

    if (ensure_data(ws,idx+min_size)==EBROK)
      return EBROK;

    if (avoid_str!=NULL) {
      if (!strncmp(ws->data_start+idx,avoid_str,avoid_size)) {
	printf("avoid found at idx=%d\n",idx);
	return EAVOID;
      }
    }
    
    if (!strncmp(ws->data_start+idx,term_str,min_size)) {
      ws->data_start+=idx;
      ws->data_size-=idx;
      if (ws->debug) {
        printf("found at idx=%d. After forward : \n",idx);
        print_ribbon(ws);
      }
      return idx;
    }

    idx++;
  } while(1);
  //never reached
  return 0;
}

int read_until_or(struct conv_ws *ws,char **result,char term1,char term2,char avoid) {
  int idx=0;
  if (ws->debug) {
    printf("read_until ");
    print_char(term1);
    printf(" or ");
    print_char(term2);
    printf(" avoid ");
    print_char(avoid);
    printf(" on \n");
    print_ribbon(ws);
  }
  do {

    if (ensure_data(ws,1)==EBROK)
      return EBROK;

    if (ws->data_start[idx]==avoid)
      return EAVOID;

    if (ws->data_start[idx]==term1 && ws->debug) {
      printf("found term1 ");
      print_char(term1);
      printf("\n");
    }
    if (ws->data_start[idx]==term2 && ws->debug) {
      printf("found term2 ");
      print_char(term2);
      printf("\n");
    }
    if (ws->data_start[idx]==term1 || ws->data_start[idx]==term2) {
      *result=(char *)malloc(idx+1);
      memcpy(*result,ws->data_start,idx);
      (*result)[idx]=0;
      ws->data_start+=idx;
      ws->data_size-=idx;
      return idx;
    }

    idx++;
    
  } while(1);
  return 0;
}
