/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include "cmd_dp.hpp"

// THREADS

void *runner(void *appli) {
  TApplication *app=(TApplication *)appli;
  //run the TApp
  app->Run();
  return NULL;
}

void *updater(void *workspace) {
  struct updater_ws *uws=(struct updater_ws*)workspace;
  while (1) {
    pthread_mutex_lock(uws->updater_mutex);
    uws->canvas->Update();
    pthread_mutex_unlock(uws->updater_mutex);
    usleep(50000); // 50000us = ~20Hz
    //usleep(33333); // 33333us = ~30Hz
    //usleep(1000000); // 1000000us = 1Hz
  }
  return NULL;
}

// EVENT LOOP FUNCTIONS

void new_block(void *plot,struct block *b) {
  DPGraph *p=(DPGraph*)plot;
  p->newblock(b);
}

void new_evt(void *plot,struct event *e) {
  DPGraph *p=(DPGraph*)plot;
  p->newevt(e);
}

void end_block(void *plot,struct block *b) {
}

void reinit(void *plot) {
}

// DPGraph CLASS

DPGraph::DPGraph(DataPlotter *dp_ws,char *ds_name,int id,char *title,char *chanx,char *chany,char *chanz,char *type,char *color) {
  this->dp_ws=dp_ws;
  this->id=id;
  this->title=title;
  this->title_len=strlen(title);
  this->chanx=chanx;
  this->chany=chany;
  this->chanz=chanz;
  first=1;
  graph=NULL;
  histo=NULL;
  graph2d=NULL;
  label=NULL;
  //launch the event loop
  loop=new_event_loop(ds_name,(void *)this,new_evt,new_block,end_block,reinit,'f');
  if (!strcmp(type,"histo")) {
    histo=new TH1F("histo","data",50,0,0);
    histo->SetTitle(title);
    histo->SetCanExtend(TH1::kAllAxes);
    gtype=HISTO;
  } else if (!strcmp(type,"3D")) {
    graph2d=new TGraph2D();
    graph2d->SetTitle(title);
    graph2d->SetLineWidth(0);
    graph2d->SetMarkerColor(atoi(color));
    graph2d->SetMarkerStyle(20);//empty circle=24 ; filled circle=20
    gtype=D3;
  } else if (!strcmp(type,"label")) {
    title=(char*)realloc(title,title_len+5);
    title[title_len]=':';
    title[title_len+1]=' ';
    title[title_len+2]=0;
    label=new TText(0.5,0.5,title);
    label->SetTextAlign(kHAlignCenter|kVAlignCenter);
    gtype=LABEL;
  } else {
    graph=new TGraph();
    graph->SetTitle(title);
    graph->SetLineWidth(1);
    graph->SetMarkerColor(atoi(color));
    graph->SetMarkerStyle(7);//empty circle=24 ; filled circle=20
    //graph->SetMarkerSize(1); //not applyable to styles 1,6,7
    gtype=GRAPH;
  }
}

DPGraph::~DPGraph() {
  for (std::vector<Filter*>::iterator f=filters.begin();f!=filters.end();f++) {
    delete (*f);
  }
  if (gtype==HISTO) {
    delete histo;
  } else if (gtype==D3) {
    delete graph2d;
  } else if (gtype==LABEL) {
    delete label;
  } else {
    delete graph;
  }
  free(title);
  free(chanx);
  free(chany);
  free(chanz);
  if (loop!=NULL) {
    free_event_loop(loop);
  }
}

void DPGraph::newblock(struct block *b) {
}

void DPGraph::newevt(struct event *e) {
  char *x_s,*y_s,*z_s;
  float x,y,z;
  char *label_content;
  printf("plot %d (pad %d)\n",id,pad);
  //determine cut with filters
  short cut=0;
  for (std::vector<Filter*>::iterator f=filters.begin();f!=filters.end();f++) {
    //printf("determining cut with filter:\n");
    //(*f)->print();
    if (!(*f)->checkEvent(e)) {
      //printf("-> cutting\n");
      cut=1;
      break;
    }
    //printf("-> not cutting\n");
  }
  if (cut) {
    return;
  }
  if (gtype==GRAPH) {
    //get data
    x_s=get_event_field(e,chanx);
    y_s=get_event_field(e,chany);
    if (x_s==NULL || y_s==NULL ||
        x_s[0]=='u' || y_s[0]=='u') {
      printf("discarding point\n");
      // if not found -> discard
      // if undef -> discard
      return;
    }
    x=atof(x_s);
    y=atof(y_s);
    //determine limits
    if (first) {
      xmin=x;
      xmax=x;
      ymin=y;
      ymax=y;
      first=0;
    } else {
      xmin=std::min(x,xmin);
      xmax=std::max(x,xmax);
      ymin=std::min(y,ymin);
      ymax=std::max(y,ymax);
    }
    //plot
    pthread_mutex_lock(dp_ws->updater_mutex);
    dp_ws->canvas->cd(pad);
    printf("plot %d: %s=%.3e and %s=%.3e\n",id,chanx,x,chany,y);
    graph->SetPoint(graph->GetN(),x,y);
    //update axis ranges
    //printf("plot %d: setting xmin=%.3e xmax=%.3e\n",id,xmin,xmax);
    graph->GetXaxis()->SetLimits(xmin,xmax);
    //printf("plot %d: setting ymin=%.3e ymax=%.3e\n",id,ymin,ymax);
    graph->SetMinimum(ymin);
    graph->SetMaximum(ymax);
    dp_ws->canvas->GetPad(pad)->Modified();
    pthread_mutex_unlock(dp_ws->updater_mutex);
  }
  if (gtype==D3) {
    //get data
    x_s=get_event_field(e,chanx);
    y_s=get_event_field(e,chany);
    z_s=get_event_field(e,chanz);
    if (x_s==NULL || y_s==NULL || z_s==NULL ||
        x_s[0]=='u' || y_s[0]=='u' || z_s[0]=='u') {
      // if not found -> discard
      // if undef -> discard
      return;
    }
    x=atof(x_s);
    y=atof(y_s);
    z=atof(z_s);
    //determine limits
    if (first) {
      xmin=x;
      xmax=x;
      ymin=y;
      ymax=y;
      zmin=z;
      zmax=z;
      first=0;
    } else {
      xmin=std::min(x,xmin);
      xmax=std::max(x,xmax);
      ymin=std::min(y,ymin);
      ymax=std::max(y,ymax);
      zmin=std::min(z,zmin);
      zmax=std::max(z,zmax);
    }
    //plot
    pthread_mutex_lock(dp_ws->updater_mutex);
    dp_ws->canvas->cd(pad);
    printf("plot %d: %s=%.3e, %s=%.3e, %s=%.3e\n",id,chanx,x,chany,y,chanz,z);
    graph2d->SetPoint(graph2d->GetN(),x,y,z);
    //update axis ranges
    /*printf("plot %d: setting xmin=%.3e xmax=%.3e\n",id,gc->xmin,gc->xmax);
    graph2d->GetXaxis()->SetLimits(gc->xmin,gc->xmax);
    printf("plot %d: setting ymin=%.3e ymax=%.3e\n",id,gc->ymin,gc->ymax);
    graph2d->GetYaxis()->SetLimits(gc->ymin,gc->ymax);
    printf("plot %d: setting zmin=%.3e zmax=%.3e\n",id,gc->zmin,gc->zmax);
    graph2d->GetZaxis()->SetLimits(gc->zmin,gc->zmax);
    */
    dp_ws->canvas->GetPad(pad)->Modified();
    pthread_mutex_unlock(dp_ws->updater_mutex);
  }
  if (gtype==HISTO) {
    //get data
    x_s=get_event_field(e,chanx);
    if (x_s==NULL || x_s[0]=='u') {
      // if not found -> discard
      // if undef -> discard
      return;
    }
    x=atof(x_s);
    pthread_mutex_lock(dp_ws->updater_mutex);
    dp_ws->canvas->cd(pad);
    histo->Fill(x);
    dp_ws->canvas->GetPad(pad)->Modified();
    pthread_mutex_unlock(dp_ws->updater_mutex);
  }
  if (gtype==LABEL) {
    //get data
    x_s=get_event_field(e,chanx);
    if (x_s==NULL || x_s[0]=='u') {
      // if not found -> discard
      // if undef -> discard
      return;
    }
    title_len=strlen(title);
    label_content=(char*)malloc(title_len+strlen(x_s)+5);
    strcpy(label_content,title);
    strcpy(label_content+title_len,x_s);
    pthread_mutex_lock(dp_ws->updater_mutex);
    dp_ws->canvas->cd(pad);
    dp_ws->canvas->GetPad(pad)->Clear();
    delete label;
    label=new TText(0.5,0.5,label_content);
    label->SetTextAlign(kHAlignCenter|kVAlignCenter);
    label->Draw();
    dp_ws->canvas->GetPad(pad)->Modified();
    pthread_mutex_unlock(dp_ws->updater_mutex);
  }
}

// DataPlotter CLASS

DataPlotter::DataPlotter() {
  nb_plots=0;

  int argc=0;
  char **argv=NULL;
  struct updater_ws *uws;

  //Root Application
  app=new TApplication("app",&argc,argv);
  canvas=new TCanvas("dataplot","Dataplot canvas",600,900);

  uws=(struct updater_ws*)malloc(sizeof(struct updater_ws));
  uws->canvas=canvas;
  uws->updater_mutex=(pthread_mutex_t*)malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(uws->updater_mutex,NULL);
  updater_mutex=uws->updater_mutex;
  updater_t=new TThread("updater_thread",updater,(void*)uws);
  if (updater_t==NULL) {
    printf("error starting updater thread\n");
    exit(1);
  }
  updater_t->Run();
  
//  serv = new THttpServer("http:8080");

  //launch the TApplication
  app_t=(pthread_t*)malloc(sizeof(pthread_t));
  if (pthread_create(app_t,NULL,runner,(void *)app)<0) {
    perror("pthread_create");
    exit(1);
  }
  pthread_detach(*app_t);
}

void DataPlotter::remap_pads() {
  //Remap plots in pads
  //This function must be called with updater_mutex locked
  canvas->Clear();
  //printf("resizing canvas to %dx%d\n",600,400*nb_plots);
  //canvas->SetCanvasSize(600,400*nb_plots);
  printf("dividing canvas in %d\n",nb_plots);
  canvas->Divide(1,nb_plots);
  int pad=1;
  for (std::map<int,DPGraph*>::iterator it=plots.begin();it!=plots.end();it++) {
    int id=it->first;
    DPGraph *p=it->second;
    p->pad=pad;
    canvas->cd(pad);
    printf("mapping plot %d to pad %d\n",id,pad);
    if (p->gtype==GRAPH) {
      p->graph->Draw("apl");
    }
    if (p->gtype==D3) {
      p->graph2d->Draw("p0");
    }
    if (p->gtype==HISTO) {
      p->histo->Draw();
    }
    if (p->gtype==LABEL) {
      p->label->Draw();
    }
    pad++;
  }
}

int DataPlotter::new_plot(char *ds_name,char *title,char *chanx,char *chany,char *chanz,char *type,char *color) {
  //Determine new id
  int id=0;
  if (plots.size()!=0) {
    id=plots.rbegin()->first+1;
  }
  //Add plot to list
  DPGraph *p=new DPGraph(this,ds_name,id,strdup(title),strdup(chanx),strdup(chany),strdup(chanz),type,color);
  if (!p || p->loop==NULL) {
    delete p;
    return -1;
  }
  pthread_mutex_lock(updater_mutex);
  plots[id]=p;
  printf("adding new plot %s (%d) with ds=%s, x=%s, y=%s, z=%s\n",title,id,ds_name,chanx,chany,chanz);
  nb_plots++;
  //Remap pads
  remap_pads();
  pthread_mutex_unlock(updater_mutex);
  return id;
}

int DataPlotter::del_plot(char *id_s) {
  //Find plot
  int id=atoi(id_s);
  std::map<int,DPGraph*>::iterator it=plots.find(id);
  if (it==plots.end()) {
    return 0;
  }
  DPGraph *p=it->second;
  pthread_mutex_lock(updater_mutex);
  delete p;
  plots.erase(id);
  nb_plots--;
  remap_pads();
  pthread_mutex_unlock(updater_mutex);
  return 1;
}

int DataPlotter::add_filter(char *id_s,char *filter_s) {
  //Find destination plot
  int id=atoi(id_s);
  std::map<int,DPGraph*>::iterator it=plots.find(id);
  if (it==plots.end()) {
    return -1;
  }
  //Get destination plot
  DPGraph *p=it->second;
  //Parse filter
  Filter *f=new Filter(filter_s);
  if (!f->valid()) {
    return -2;
  }
  //Add plot
  printf("adding filter to plot %d:\n",id);
  f->print();
  p->filters.push_back(f);
  return 0;
}

int DataPlotter::del_filters(char *id_s) {
  //Find destination plot
  int id=atoi(id_s);
  std::map<int,DPGraph*>::iterator it=plots.find(id);
  if (it==plots.end()) {
    return -1;
  }
  //Get destination plot
  DPGraph *p=it->second;
  //Delete filters
  for (std::vector<Filter*>::iterator it=p->filters.begin();it!=p->filters.end();it++) {
    delete (*it);
    p->filters.erase(it);
  }
  return 0;
}

void DataPlotter::clrscr(DPGraph *p) {
  p->first=1;
  if (p->gtype==GRAPH) {
    p->graph->Set(0);
  }
  if (p->gtype==D3) {
    p->graph2d->Set(0);
  }
  if (p->gtype==HISTO) {
    p->histo->Reset();
  }
  if (p->gtype==LABEL) {
    canvas->GetPad(p->pad)->Clear();
  }
}

void DataPlotter::clrscrall() {
  for (std::map<int,DPGraph*>::iterator it=plots.begin(); it!=plots.end(); it++) {
    clrscr(it->second);
  }
}

void DataPlotter::rewind() {
  for (std::map<int,DPGraph*>::iterator it=plots.begin(); it!=plots.end(); it++) {
    DPGraph *p=it->second;
    int mode=p->loop->mode;
    stop_loop(p->loop);
    rewind_loop(p->loop);
    clrscrall();
    clean_loop(p->loop);
    if (mode==RUNNING) {
      run_loop(p->loop);
    }
  }
}

void DataPlotter::clean() {
  for (std::map<int,DPGraph*>::iterator it=plots.begin(); it!=plots.end(); it++) {
    DPGraph *p=it->second;
    int mode=p->loop->mode;
    stop_loop(p->loop); 
    clrscrall();
    if (mode==RUNNING) {
      run_loop(p->loop);
    }
  }
}

char *DataPlotter::list_plots() {
  short first=1;
  struct incr_string *result=new_is();
  char *plot_name;
  for (std::map<int,DPGraph*>::iterator it=plots.begin();it!=plots.end();it++) {
    DPGraph *p=it->second;
    plot_name=strdup(p->title);
    if (p->gtype==LABEL) {
      // remove ": "
      plot_name[strlen(p->title)-2]=0;
    }
    if (first) {
      fsuffix_is(result,"%s",plot_name);
      first=0;
    } else {
      fsuffix_is(result,",%s",plot_name);
    }
    free(plot_name);
  }
  return reduce_is(result);
}

void DataPlotter::stop() {
  for (std::map<int,DPGraph*>::iterator it=plots.begin(); it!=plots.end(); it++) {
    DPGraph *p=it->second;
    stop_loop(p->loop);
  }
}

void DataPlotter::onestep() {
  for (std::map<int,DPGraph*>::iterator it=plots.begin(); it!=plots.end(); it++) {
    DPGraph *p=it->second;
    onestep_loop(p->loop);
  }
}

void DataPlotter::run() {
  for (std::map<int,DPGraph*>::iterator it=plots.begin(); it!=plots.end(); it++) {
    DPGraph *p=it->second;
    run_loop(p->loop);
  }
}

int DataPlotter::change_policy(char *newpol) {
  char policy;
  policy=newpol[0];
  if (policy!='f' && policy!='l') {
    return 0;
  }
  for (std::map<int,DPGraph*>::iterator it=plots.begin(); it!=plots.end(); it++) {
    DPGraph *p=it->second;
    changepol_loop(p->loop,policy);
  }
  return 1;
}

// Pyrame EXTERNAL FUNCTIONS

//pyrapi:new_plot_dp:data_source,plot_name,chan_x,chan_y,chan_z,type,color
extern "C" struct cmd_result *new_plot_dp(struct cmd *,void *);
struct cmd_result *new_plot_dp(struct cmd *command,void *workspace) {
  int res;
  DataPlotter*plotter=(DataPlotter *)workspace;
  if (command->nb_params<6) {
    return build_result(0,"invalid number of arguments");
  }
  if (command->nb_params<7) {
    res=plotter->new_plot(command->params[0],command->params[1],command->params[2],command->params[3],command->params[4],command->params[5],"1");
  } else {
    res=plotter->new_plot(command->params[0],command->params[1],command->params[2],command->params[3],command->params[4],command->params[5],command->params[6]);
  }
  if (res==-1) {
    return build_result(0,"error creating event loop or invalid data source");
  }
  return build_result(1,"%d",res);
}

//pyrapi:del_plot_dp:plot_id
extern "C" struct cmd_result *del_plot_dp(struct cmd *,void *);
struct cmd_result *del_plot_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  int res=plotter->del_plot(command->params[0]);
  if (!res) {
    return build_result(0,"invalid id");
  }
  return build_result(1,"ok");
}

//pyrapi:add_filter_dp:plot_id,filter
extern "C" struct cmd_result *add_filter_dp(struct cmd *,void *);
struct cmd_result *add_filter_dp(struct cmd *command,void *workspace) {
  int res;
  DataPlotter*plotter=(DataPlotter *)workspace;
  if (command->nb_params<2) {
    return build_result(0,"invalid number of arguments");
  }
  res=plotter->add_filter(command->params[0],command->params[1]);
  if (res==-1) {
    return build_result(0,"invalid id");
  }
  if (res==-2) {
    return build_result(0,"invalid filter");
  }
  return build_result(1,"ok");
}

//pyrapi:del_filters_dp:plot_id
extern "C" struct cmd_result *del_filters_dp(struct cmd *,void *);
struct cmd_result *del_filters_dp(struct cmd *command,void *workspace) {
  int res;
  DataPlotter*plotter=(DataPlotter *)workspace;
  if (command->nb_params<1) {
    return build_result(0,"invalid number of arguments");
  }
  res=plotter->del_filters(command->params[0]);
  if (res==-1) {
    return build_result(0,"invalid id");
  }
  return build_result(1,"ok");
}

//pyrapi:change_policy_dp:policy
extern "C" struct cmd_result *change_policy_dp(struct cmd *,void *);
struct cmd_result *change_policy_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  int res=plotter->change_policy(command->params[0]);
  if (res==0) {
    return build_result(0,"unknown policy (should be \"first\" or \"last\")");
  } else {
    return build_result(1,"ok");
  }
}

//pyrapi:rewind_dp:
extern "C" struct cmd_result *rewind_dp(struct cmd *,void *);
struct cmd_result *rewind_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  plotter->rewind();
  return build_result(1,"ok");
}

//pyrapi:run_dp:
extern "C" struct cmd_result *run_dp(struct cmd *,void *);
struct cmd_result *run_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  plotter->run();
  return build_result(1,"ok");
}

//pyrapi:stop_dp:
extern "C" struct cmd_result *stop_dp(struct cmd *,void *);
struct cmd_result *stop_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  plotter->stop();
  return build_result(1,"ok");
}

//pyrapi:onestep_dp:
extern "C" struct cmd_result *onestep_dp(struct cmd *,void *);
struct cmd_result *onestep_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  plotter->onestep();
  return build_result(1,"ok");
}

//pyrapi:clean_dp:
extern "C" struct cmd_result *clean_dp(struct cmd *,void *);
struct cmd_result *clean_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  plotter->clean();
  return build_result(1,"ok");
}

//pyrapi:list_plots_dp:
extern "C" struct cmd_result *list_plots_dp(struct cmd *,void *);
struct cmd_result *list_plots_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  struct cmd_result *cres;
  char *list=plotter->list_plots();
  cres=build_result(1,list);
  free(list);
  return cres;
}

//pyrapi:save_root_dp:
extern "C" struct cmd_result *save_root_dp(struct cmd *,void *);
struct cmd_result *save_root_dp(struct cmd *command,void *workspace) {
  DataPlotter*plotter=(DataPlotter *)workspace;
  if (command->nb_params<1) {
    return raw_build_result(0,"invalid number of arguments");
  }
  plotter->canvas->SaveAs(command->params[0]);
  return raw_build_result(1,"ok");
}

extern "C" void*init(int,char **);
void *init(int argc, char **argv) {
  //return the workspace to cmdmod
  return (void *)new DataPlotter();
}

