/*
 * Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
 * This file is part of Pyrame.
 *
 * Pyrame is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pyrame is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
 * */

#include "cmd_evb.h"

void *init() {
  struct evb_workspace *ws=malloc(sizeof(struct evb_workspace));
  memset(ws,0,sizeof(struct evb_workspace));
  submod_newns("default");
  return (void *)ws;
}

//pyrapi:init_evb:dev_name,sources_name,reco_lib,choice_field,tolerance,output_pattern,cutarg1,cutarg2,cutarg3,varmod_id
struct cmd_result *init_evb(struct cmd *command,void *workspace) {
  struct evb_workspace *ws=(struct evb_workspace *)workspace;
  struct cmd_result *cres;
  struct cmd_result *es_cres;
  int i;
  char **ds_list;
  int nb_ds;
  char **param_list;
  int nb_params;
  int res;
  char port[128];
  
  if (ws->initialized!=0) {
    return build_result(1,"evb has been already been initialized");
  }

  //init stats system
  printf("init the stats system\n");
  ws->stats=init_stats("127.0.0.1",1000000);
  
  if (command->nb_params!=10) {
    return build_result(0,"init_evb, %d arguments given, 10 needed : dev_name sources_name reco_lib choice_field tolerance output_pattern cutarg1 cutarg2 cutarg3 varmod_id",command->nb_params);
  }

  //params are
  //[0] dev_name
  //[1] sources_name
  //[2] reco_lib
  //[3] choice field
  //[4] choice tolerance t is related to t0 iff t0<=t<=t0+tol
  //[5] output pattern
  //[6] cut argument 1
  //[7] cut argument 2
  //[8] cut argument 3
  //[9] varmod id

  submod_setvar("default","varmod_id",command->params[9],ENV_STR);


  //initialize the dispatcher
  sprintf(port,"%d",submod_getport("EVB_PORT"));
  res=submod_init_dispatcher(BLOCK_NB,"evb",command->params[5],port);
  if (res<0) {
    return build_result(0,"cant init evb : cant init dispatcher : %s",pyrame_errors[-res]);
  }

  //get the data source ids
  //sleep(1);
  cres=submod_execcmd("find_ds@varmod",command->params[9],"default",command->params[1],"end");
  if (cres->retcode==0) {
    return build_error(cres,"cant get data sources : %s");
  }

  //printf("result from varmod=%d,%s\n",cres->retcode,cres->str);
  ds_list=split(cres->str,',',&nb_ds);
  //printf("nb_ds=%d\n",nb_ds);
  free_cmd_result(cres);
  if (nb_ds==0)
    return build_result(0,"get empty data sources");

  //allocate the sockets arrays and evtstruct
  ws->data_sockets=malloc(sizeof(int)*nb_ds);
  ws->dr_sockets=malloc(sizeof(int)*nb_ds);
  ws->es=malloc(sizeof(struct evtstruct *)*nb_ds);
  ws->nb_data_clients=nb_ds;
  printf("nb of data clients : %d\n",ws->nb_data_clients);
  
  for(i=0;i<nb_ds;i++) {
    cres=submod_execcmd("get_ds@varmod",command->params[9],"default",ds_list[i],"end");
    if (cres->retcode==0) {
      return build_error(cres,"cant get data source : %s");
    }
    param_list=split(cres->str,',',&nb_params);
    free_cmd_result(cres);
    ws->data_sockets[i]=open_socket(param_list[0],atoi(param_list[1]));
    if (ws->data_sockets[i]==-1)
      return build_result(0,"unable to open data socket for ds %d on %s:%s",i,param_list[0],param_list[1]);
    ws->dr_sockets[i]=open_socket(param_list[0],atoi(param_list[2]));
    if (ws->dr_sockets[i]==-1)
      return build_result(0,"unable to open data ready socket for ds %d on %s:%s",i,param_list[0],param_list[2]);
    free_split(param_list,nb_params);
    es_cres=execcmd(ws->data_sockets[i],"getchans_module","end");
    if (es_cres->retcode==0) {
      return build_error(es_cres,"cant get channels from converter <- %s");
    } else {
      ws->es[i]=saf_parse_chans(es_cres->str);
      free_cmd_result(es_cres);
    }
  }
  free_split(ds_list,nb_ds);
    

  //fill the time field params
  ws->choice_field=strdup(command->params[3]);
  ws->tolerance=atof(command->params[4]);

  //create the evtstruct for output
  ws->output_es=saf_parse_chans(command->params[5]);

  //store the cut arguments
  ws->cutarg1=strdup(command->params[6]);
  ws->cutarg2=strdup(command->params[7]);
  ws->cutarg3=strdup(command->params[8]);

  //register params in cmod
  if (strcmp(command->params[0],"undef")) {
    submod_execcmd("set_param@cmod",ws->cmod_id,"evb_sources_name",command->params[2],"end");
    submod_execcmd("set_param@cmod",ws->cmod_id,"evb_cutarg1",ws->cutarg1,"end");
    submod_execcmd("set_param@cmod",ws->cmod_id,"evb_cutarg2",ws->cutarg2,"end");
    submod_execcmd("set_param@cmod",ws->cmod_id,"evb_cutarg3",ws->cutarg3,"end");
    submod_execcmd("set_param@cmod",ws->cmod_id,"evb_reco_lib",command->params[3],"end");
    submod_execcmd("set_param@cmod",ws->cmod_id,"evb_choice_field",command->params[4],"end");
    submod_execcmd("set_param@cmod",ws->cmod_id,"evb_tolerance",command->params[5],"end");
    submod_execcmd("set_param@cmod",ws->cmod_id,"evb_output_pattern",command->params[6],"end");
  }

  submod_set_chans(command->params[5]); 
  
  //lib
  if (strcmp(command->params[2],"undef")) {
    ws->lib_phy=dlopen(command->params[2],RTLD_NOW|RTLD_GLOBAL);
    if (!ws->lib_phy) {
      printf("dlopen error : %s...exiting\n",dlerror()); 
      submod_exit(51);
    }
    ws->cut=dlsym(ws->lib_phy,"cut");
  } else {
    ws->lib_phy=NULL;
    ws->cut=NULL;
  }
  
  //start the manager thread
  printf("Starting the manager thread\n");
  if (pthread_create(&(ws->t_manager),NULL,manager,ws)<0) {
    perror("pthread_create");
    submod_exit(50);
  }
  pthread_detach(ws->t_manager);
  
  
  ws->initialized=1;
  printf("event builder successfully initialized\n");
  return build_result(1,"ok");
}

//pyrapi:deinit_evb:dev_name
struct cmd_result *deinit_evb(struct cmd *command,void *workspace) {
  struct evb_workspace *ws=(struct evb_workspace *)workspace;
  int i;
  int timeout=0;

  if (!ws->initialized) {
    return build_result(1,"evb already deinitilized");
  }
  
  //stop manager thread
  ws->command=1;
  while(ws->command!=0) {
    sleep(1);
    timeout++;
    if (timeout==3) {
      printf("evb : timeout : cant kill manager properly : killing hardly\n");
      pthread_cancel(ws->t_manager);
      break;
    }
  }

  submod_flush_dispatcher();
  ws->block_cnt=0;
  
  //close all sockets
  for (i=0;i<ws->nb_data_clients;i++) {
    close(ws->data_sockets[i]);
    close(ws->dr_sockets[i]);
    free_evtstruct(ws->es[i]);
  }

  //free all resources
  free(ws->data_sockets);
  free(ws->dr_sockets);
  
  free(ws->es);
  if (ws->lib_phy) {
    dlclose(ws->lib_phy);
  }
  free(ws->cmod_id);
  free(ws->choice_field);
  free(ws->cutarg1);
  free(ws->cutarg2);
  free(ws->cutarg3);

  //deinit
  ws->initialized=0;
  printf("deinitialized\n");
  return build_result(1,"ok");
}

//pyrapi:start_acq_evb:dev_name
struct cmd_result *start_acq_evb(struct cmd *command,void *workspace) {
  struct evb_workspace *ws=(struct evb_workspace *)workspace;
  ws->acquiring=1;
  return build_result(1,"ok");
}

//pyrapi:stop_acq_evb:dev_name
struct cmd_result *stop_acq_evb(struct cmd *command,void *workspace) {
  struct evb_workspace *ws=(struct evb_workspace *)workspace;
  ws->acquiring=0;
  return build_result(1,"ok");
}


void get_token(struct evb_workspace *ws) {
  struct cmd_result *res_vm=NULL;
  //printf("getting the token...");
  do {
    if (res_vm!=NULL)
      free_cmd_result(res_vm);
    res_vm=submod_execcmd("get_token@varmod",submod_getvar("default","varmod_id"),"default","evb_token","end");
    if (res_vm->retcode==0) {
      usleep(50000);
    }
  } while (res_vm->retcode==0);
  free_cmd_result(res_vm);
  //printf("got the token\n");
}

void release_token(struct evb_workspace *ws) {
  struct cmd_result *res_vm;
  //printf("releasing the token...");
  res_vm=submod_execcmd("release_token@varmod",submod_getvar("default","varmod_id"),"default","evb_token","end");
  free_cmd_result(res_vm);
  //printf("token released\n");
}

void update_ltr(struct evb_workspace *ws,int choosen_time) {
  struct cmd_result *res_vm;
  char choosen_time_str[1024];
  sprintf(choosen_time_str,"%d",choosen_time);
  res_vm=submod_execcmd("setvar@varmod",submod_getvar("default","varmod_id"),"default","evb_last_time",choosen_time_str,"end");
  if (res_vm->retcode==0) {
    printf("cant set the last time on varmod because %s ... exiting\n",res_vm->str);
    submod_exit(57);
  }
  free_cmd_result(res_vm);
}

int get_ltr(struct evb_workspace *ws) {
  int ltr;
  struct cmd_result *res_vm;
  res_vm=submod_execcmd("getvar@varmod",submod_getvar("default","varmod_id"),"default","evb_last_time","end");
  if (res_vm->retcode==0) {
    //the variable does not exists -> creating it with value 0
    printf("warning : unable to get lrt -> creating it!!!!! : %s\n",res_vm->str);
    update_ltr(ws,0);
    free_cmd_result(res_vm);
    return 0;
  } else {
    ltr=atoi(res_vm->str);
    free_cmd_result(res_vm);
    return ltr;
  }
}

void *manager(void *args)
{
  int i,j,k;
  struct evb_workspace *ws=(struct evb_workspace *)args;
  int last_time;
  int choosen_time;
  struct id_list **time_ranges;
  fd_set readfds;
  int highest=0;
  char trash[1024];
  int res;
  struct timeval tv;
  //int data_nb;
  int neigh_cnt;
  struct block *b;
  struct timeval *last_full;
  char *stuck;
  int timeout=2000000;
  struct evtc *container;
  float smallest;
  struct event *e;
  unsigned int high_limit;
  
  printf("waiting for initialization\n");
  while(ws->initialized==0) {
    usleep(10000);
  }

  //allocate the time ranges array
  time_ranges=malloc(sizeof(struct id_list*)*ws->nb_data_clients);
  for(i=0;i<ws->nb_data_clients;i++) {
    time_ranges[i]=NULL;
  }

  //allocate timeout system
  last_full=malloc(sizeof(struct timeval)*ws->nb_data_clients);
  for(i=0;i<ws->nb_data_clients;i++)
    gettimeofday(&last_full[i],NULL);
  stuck=malloc(sizeof(char)*ws->nb_data_clients);
  memset(stuck,0,sizeof(char)*ws->nb_data_clients);
  
  printf("initialized\n");

  //release that if you are working with multiple evb
  release_token(ws);
  update_ltr(ws,0);

  //event building
  while(1) {

    if (ws->command==1) {
      break;
    }
    
    if (ws->acquiring==0) {
      usleep(10000);
      for(i=0;i<ws->nb_data_clients;i++)
        gettimeofday(&last_full[i],NULL);
      continue;
    }

    //step 1 get the token from varmod
    //printf("step 1 : getting token\n");
    get_token(ws);
    
    //step 2 get the last_time_range from varmod
    //printf("step 2 : getting last spill from varmod\n");
    last_time=get_ltr(ws);
    
    
    //step 3 ask every data their smaller time beyond last_time_range
    //printf("step 3 : getting available spills\n");
    high_limit=(unsigned int)(-1);

    for (i=0;i<ws->nb_data_clients;i++) {
      time_ranges[i]=get_block_list(ws->data_sockets[i],last_time);
      
      if (time_ranges[i]==NULL) {
        printf("no converter/dispatcher reachable for socket %d...exiting\n",i);
        submod_exit(58);
      }
      //printf("client=%d\n",i);
      //print_id_list(time_ranges[i]);

      //find the lowest of the highest
      gettimeofday(&tv,NULL);
      if (time_ranges[i]->nb_ids>=1) { //client has data
        gettimeofday(&last_full[i],NULL);
        if (stuck[i]!=0) {
          printf("warning : source %d is unstuck\n",i);
          stuck[i]=0;
        }
        if (time_ranges[i]->ids[time_ranges[i]->nb_ids-1]<high_limit) {
          high_limit=time_ranges[i]->ids[time_ranges[i]->nb_ids-1];
          printf("high_limit for %d becomes %d\n",i,high_limit);
        }
      } else { //client has no data
        if (stuck[i]!=1) { //client is not stuck
          if ((tv.tv_sec*1000000+tv.tv_usec)>(last_full[i].tv_sec*1000000+last_full[i].tv_usec+timeout)) { //if timeout is exceeded
            stuck[i]=1;
            printf("warning : source %d is stuck\n",i);
          } else { //timeout is not exceeded
            high_limit=last_time;
            printf("client %d has no data and is not stuck -> going to last time\n",i);
          }
        }
      }
    }
    printf("last_time=%d high_limit=%d\n",last_time,high_limit);


    //step 4 choose the best time
    //printf("step 4 : choosing the best spill\n");
    //parcour the first answer and find the lowest number of it that is in all other answers

    choosen_time=-1;
    if (high_limit!=(unsigned int)(-1)) {
      for(i=last_time+1;i<=high_limit;i++) {
	neigh_cnt=0;
        printf("count neighbour for %d\n",i);
	for(j=0;j<ws->nb_data_clients;j++) {
	  for (k=0;k<time_ranges[j]->nb_ids;k++) {
	    if (time_ranges[j]->ids[k]==i) {
	      neigh_cnt++;
	      break;
	    }
	  }
	}
        printf("neigh_cnt=%d\n",neigh_cnt);
	if (neigh_cnt>=2) {
	  choosen_time=i;
	  break;
	} 
      }
    }
    printf("choosen time=%d\n",choosen_time);

    //free the bl
    for (i=0;i<ws->nb_data_clients;i++) {
      if (time_ranges[i]) {
        free_id_list(time_ranges[i]);
        time_ranges[i]=NULL;
      }
    }

    
    
    //if no time range is choosen -> release the token, wait for new data and continue the loop
    if (choosen_time==-1) {
      printf("no time range available\n");
      release_token(ws);
      //waiting for any new data available
      FD_ZERO(&readfds);
      highest=-1;
      for (i=0;i<ws->nb_data_clients;i++) {
        if (ws->dr_sockets[i]!=-1) {
          FD_SET(ws->dr_sockets[i],&readfds);
          //printf("adding %d\n",ws->dr_sockets[i]);
          if (ws->dr_sockets[i]>highest) {
            highest=ws->dr_sockets[i];
          }
        }
      }
      if (highest==-1) {
        printf("no socket to add, skipping select and waiting 1s\n");
        sleep(1);
      } else {
        //printf("go to select\n");
        tv.tv_sec=1;
        tv.tv_usec=0;
        select(highest+1,&readfds,NULL,NULL,&tv);
        for (i=0;i<ws->nb_data_clients;i++) {
          if (FD_ISSET(ws->dr_sockets[i],&readfds)) {
            res=read(ws->dr_sockets[i],trash,1024);
            if (res<=0) {
              printf("warning: lost data ready socket\n");
            }
          }
        }
      }
      continue; //main loop
    }
    
    //step 5 update the last_time_range on varmod
    //printf("step 5 : update the choosen spill in varmod\n");
    update_ltr(ws,choosen_time);

    //step 6 give back the token
    //printf("step 6 : release token\n");
    release_token(ws);
    
    //step 7 create events container
    //printf("step 7 : create an event container\n");
    container=new_evtc();
    
    //step 8 get data in lle pool
    //printf("step 8 :getting data\n");
    for (i=0;i<ws->nb_data_clients;i++) {
      //printf("getting data for socket %d\n",i);
      b=pull_block(ws->data_sockets[i],choosen_time,ws->es[i]);
      if (b!=NULL)
	add_block_evtc(container,b);
    }
    
    //step 9 apply builder
    //printf("step 9 : building\n");
    smallest=find_smallest_evtc(container,ws->choice_field);
    while (smallest!=-1.0) {
      printf("found smallest=%f\n",smallest);
      //printf("new block\n");
      b=new_block_from_es(ws->output_es,0);
      e=find_related_evtc(container,ws->choice_field,smallest,ws->tolerance);
      while(e!=NULL) {
        //printf("found a new event\n");
        //print_event(e);
        convert_from_evtstruct(e,ws->output_es,b);
        e=find_related_evtc(container,ws->choice_field,smallest,ws->tolerance);
      }
      //cut the block if necessary
      if ((!ws->cut) || (ws->cut(b,ws->cutarg1,ws->cutarg2,ws->cutarg3)==0)) {
        //dump the block
        //printf("dumping block with id %d: \n",ws->block_cnt);
        //print_block(b);
        submod_add_block(b,ws->block_cnt);
        ws->block_cnt++;
      } else {
        free_block(b);
      }

      //printf("find new smallest\n");
      smallest=find_smallest_evtc(container,ws->choice_field);
    } 
    free_evtc(container);

    //step 10 waiting for user
    //printf("Press any key to continue\n");
    //getchar();
    
  } //while(1)

  free(time_ranges);
  free(last_full);
  free(stuck);
  ws->command=0;
  return NULL;
}

