/*
 * Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
 * This file is part of Pyrame.
 *
 * Pyrame is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pyrame is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
 * */

#include <pyrame.h>

char wait_for_data(struct evloop *ws) {
  struct timeval tv;
  fd_set readfds;
  char trash[1024];
  int res;
  int ca=0;

  if (ws->dr_socket==-1) {
    ws->dr_socket=open_socket(ws->host,ws->datareadyport);
    if (ws->dr_socket==-1) {
      printf("unable to open data ready socket\n");
      return 0;
    }
  }

  tv.tv_sec=2;
  tv.tv_usec=000000;
  FD_ZERO(&readfds);
  FD_SET(ws->dr_socket,&readfds);
  ca=select(ws->dr_socket+1,&readfds,NULL,NULL,&tv);
  
  if (ca==0) {
    //timeout
    return 0;
  }
  
  res=read(ws->dr_socket,trash,1);
  if (res<=0) {
    printf("warning : eventloop : control socket lost\n");
    close(ws->dr_socket);
    ws->dr_socket=-1;
    return 0;
  }

  close(ws->dr_socket);
  ws->dr_socket=-1;
  return 1;
}

void rewind_loop(struct evloop *ws) {
  ws->command=CMD_REWIND;
  while(ws->command!=0)
    usleep(CMD_SLEEP_TIME);
}

void onestep_loop(struct evloop *ws) {
  ws->command=CMD_ONESTEP;
  while(ws->command!=0)
    usleep(CMD_SLEEP_TIME);
}

void oneblock_loop(struct evloop *ws) {
  ws->command=CMD_ONEBLOCK;
  while(ws->command!=0)
    usleep(CMD_SLEEP_TIME);
}

void stop_loop(struct evloop *ws) {
  ws->command=CMD_STOP;
  while(ws->command!=0)
    usleep(CMD_SLEEP_TIME);
}

void clean_loop(struct evloop *ws) {
  ws->command=CMD_CLEAN;
  while(ws->command!=0)
    usleep(CMD_SLEEP_TIME);
}

void run_loop(struct evloop *ws) {
  printf("going to run mode\n");
  ws->command=CMD_RUN;
  while(ws->command!=0)
    usleep(CMD_SLEEP_TIME);
  if (ws->mode!=RUNNING) {
    printf("Error : event loop not in run mode (%d should be %d)\n",ws->mode,RUNNING);
  } else {
    printf("event loop in run mode\n");
  }
}

void changepol_loop(struct evloop *ws,char new_policy) {
  ws->new_pol=new_policy;
  ws->command=CMD_CHANGE_POL;
  while(ws->command!=0)
    usleep(CMD_SLEEP_TIME);
}

        
void *evloop(void *workspace) {
  struct evloop *ws=(struct evloop *)workspace;
  int new_key;
  int result;
  int socket=-1;
  struct cmd_result *cres;
  ws->b=NULL;
  ws->e=NULL;
  socket=open_socket(ws->host,ws->dataport);

  while (1) {

    //exec command
    if (ws->command!=0) {
      //printf("executing command %d\n",ws->command);
      switch(ws->command) {
      case CMD_CLEAN:
        break;
      case CMD_REWIND:
        ws->last_key=0;
        break;
      case CMD_STOP:
        ws->mode=STOPPED;
        break;
      case CMD_RUN:
        ws->mode=RUNNING;
        break;
      case CMD_ONESTEP:
        ws->mode=ONESTEP;
        break;
      case CMD_ONEBLOCK:
        ws->mode=ONEBLOCK;
        break;
      case CMD_CHANGE_POL:
        ws->policy=ws->new_pol;
        break;
      }
      ws->command=0;
    }

    if (ws->mode==STOPPED) {
      //printf("mode stopped...waiting for start\n");
      usleep(50000);
      continue;
    }

    if (socket==-1) {
      printf("try to reopen the data socket %s:%d\n",ws->host,ws->dataport);
      socket=open_socket(ws->host,ws->dataport);
      continue;
    }

    if (ws->es==NULL) {
      cres=execcmd(socket,"getchans_module","end");
      if (cres->retcode==0) {
        printf("cant get channels from data source\n");
        free_cmd_result(cres);
        continue;
      } else {
        ws->es=saf_parse_chans(cres->str);
        if (ws->es==NULL) {
          printf("error: cant parse schema: %s\n",cres->str);
        }
        free_cmd_result(cres);
      }
    }

    
    if (ws->e==NULL) {
      //no more event available
      
      //freeing the old block
      if (ws->b) {
        ws->endblock_func(ws->workspace,ws->b);
        free_block(ws->b);
        ws->b=NULL;
      }

      //stop if necessary
      if (ws->mode==ONEBLOCK) {
        ws->mode=STOPPED;
      }

      
      //choose the block depending on the policy
      result=smart_get_block(socket,ws->policy,ws->last_key,&ws->run_nb,&new_key,&(ws->b),ws->es);
      if (result!=1) {
        switch (result) {
          case -1:
            //printf("el: waiting for data (last_key=%d)\n",ws->last_key);
            wait_for_data(ws);
            break;
          case -2:
            printf("event loop: starting new run %d\n",ws->run_nb);
            ws->reinit_func(ws->workspace);
            break;
          case -3:
            printf("event loop: warning: bad socket\n");
            break;
          default:
            break;
        }
        continue;
      }

      //complete the block values
      ws->b->id=new_key;
      ws->b->es=ws->es;
      ws->last_key=new_key;

      //call the new block function
      ws->newblock_func(ws->workspace,ws->b);
      ws->e=ws->b->events;

      //treat command before treating next event
      continue;
    } else {
      //treating event
      ws->newevt_func(ws->workspace,ws->e);
      ws->e=ws->e->next;
      if (ws->mode==ONESTEP) {
        printf("one step\n");
        ws->mode=STOPPED;
      }
    }
  } 
}


struct evloop * new_event_loop(char * ds_name,void *workspace,void (*newevt_func)(void *,struct event *),void (*newblock_func)(void *,struct block *),void (*endblock_func)(void *,struct block *),void (*reinit_func)(void *),char policy) {
  struct cmd_result *pres;
  struct evloop *result=malloc(sizeof(struct evloop));
  int nbres;
  char **splres;
  result->command=0;
  result->ds_name=ds_name;

  //TODO : add a real address
  pres=sendcmd("127.0.0.1",get_port("VARMOD_PORT"),"get_ds_varmod","varmod_id","default",ds_name,"end");
  if (pres->retcode==0) {
    printf("error : unable to find data source %s\n",ds_name);
    free_cmd_result(pres);
    free(result);
    return NULL;
  }
  splres=split(pres->str,',',&nbres);
  result->host=strdup(splres[0]);
  result->dataport=atoi(splres[1]);
  result->datareadyport=atoi(splres[2]);
  free_split(splres,nbres);
  free_cmd_result(pres);
  result->workspace=workspace;
  result->mode=STOPPED;
  result->newevt_func=newevt_func;
  result->newblock_func=newblock_func;
  result->endblock_func=endblock_func;
  result->reinit_func=reinit_func;
  result->es=NULL;
  result->avail_data=0;
  result->run_nb=0;
  result->policy=policy;
  result->dr_socket=-1;
  result->last_key=-1;
  result->t_loop=(pthread_t*)malloc(sizeof(pthread_t));
  if (pthread_create(result->t_loop,NULL,evloop,(void *)result)<0) {
    perror("pthread_create...exiting");
    exit(3);
  }
  pthread_detach(*result->t_loop);
  return result;
}
  
void free_event_loop(struct evloop *el) {
  pthread_cancel(*el->t_loop);
  sleep(0.1);
  free(el->t_loop);
  free(el);
}
