/*
 * Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
 * This file is part of Pyrame.
 *
 * Pyrame is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pyrame is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
 * */

#include <pyrame.h>

void print_field(enum field domain) {
  switch(domain) {
  case PROPS:
    printf("props");
    break;
  case TIME:
    printf("time");
    break;
  case SPACE:
    printf("space");
    break;
  case DATA:
    printf("data");
    break;
  }
}

void print_evtstruct(struct evtstruct *es) {
  struct evtstructelem *ese;
  if (es==NULL) {
    printf("null evtstruct\n");
    return;
  }
  ese=es->first;
  while(ese!=NULL) {
    printf("name=%s domain=",ese->name);
    print_field(ese->domain);
    printf(" index=%d\n",ese->index);
    ese=ese->next;
  }
}

void free_evtstruct(struct evtstruct *es) {
  struct evtstructelem *ese;
  struct evtstructelem *todelete;
  if (es==NULL) {
    return;
  }
  ese=es->first;
  while(ese!=NULL) {
    todelete=ese;
    ese=ese->next;
    free(todelete);
  }
  free(es);
}

char *get_block_field(struct block *b,char *field_name) {
  struct evtstructelem *ese;
  if ((b==NULL) || (b->es==NULL))
    return NULL;
  ese=get_ese(b->es,field_name);
  if (ese!=NULL) {
    if (ese->index<b->props_size)
      return b->props[ese->index];
    else
      return NULL;
  }
  printf("warning : get_block_field : field %s not found\n",field_name);
  return NULL;
}

int set_block_field(struct block *b,struct evtstructelem *ese,char *value) {
  if (b==NULL) {
      printf("error : block is null\n");
    return -EBLNULL;
  }

  if (b->es==NULL) {
    printf("error : block es is null\n");
    return -EBLNULL;
  }
  
  if (b->props[ese->index]!=NULL) {
    free(b->props[ese->index]);
  }
  b->props[ese->index]=value;
  return 0;
}

char *get_event_field(struct event *e,char *field_name) {
  struct evtstructelem *ese;
  if ((e==NULL) || (e->block==NULL) || (e->block->es==NULL))
    return NULL;
  ese=get_ese(e->block->es,field_name);
  if (ese!=NULL) {
    switch(ese->domain) {
    case PROPS:
      if (e->block==NULL) {
        printf("error : no block set for this event\n");
        return NULL;
      }
      if (ese->index<e->block->props_size)
        return e->block->props[ese->index];
      else
        return NULL;
    case TIME:
      if (ese->index<e->time_size)
        return e->time[ese->index];
      else
        return NULL;
    case SPACE:
      if (ese->index<e->space_size)
        return e->space[ese->index];
      else
        return NULL;
    case DATA:
      if (ese->index<e->data_size)
        return e->data[ese->index];
      else
        return NULL;
    }
  }
  //not found
  printf("warning : get_event_field : field %s not found\n",field_name);
  return NULL;
}

struct evtstructelem *get_ese(struct evtstruct *es,char *field_name) {
  struct evtstructelem *ese;
  ese=es->first;
  while(ese!=NULL) {
    if (!strcmp(field_name,ese->name)) {
      return ese;
    }
    ese=ese->next;
  }
  //not found
  return NULL;
}

int set_event_field(struct event *e,struct evtstructelem *ese,char *value) {
  char not_found=1;

  if ((e==NULL) || (e->block==NULL) || (e->block->es==NULL))
    return -EBLNULL;
  
  switch(ese->domain) {
  case TIME:
    if (e->time[ese->index]!=NULL) {
      free(e->time[ese->index]);
    }
    e->time[ese->index]=value;
    not_found=0;
    break;
  case SPACE:
    if (e->space[ese->index]!=NULL) {
      free(e->space[ese->index]);
    }
    e->space[ese->index]=value;
    not_found=0;
    break;
  case DATA:
   if (e->data[ese->index]!=NULL) {
      free(e->data[ese->index]);
    }
    e->data[ese->index]=value;
    not_found=0;
    break;
  default:
    break;
  }
  //not found
  if (not_found) {
    return -ENOFIELD;
  }
  return 0;
}


//struct evtstructelem {
//  char *name;
//  enum field domain;
//  int index;
//  struct evtstructelem *next;
//};
void convert_from_evtstruct(struct event *source,struct evtstruct *convert_es,struct block *b) {
  struct event *e=new_event(b);
  e->time=malloc(sizeof(char *)*convert_es->nb_time);
  e->time_size=convert_es->nb_time;
  e->space=malloc(sizeof(char *)*convert_es->nb_space);
  e->space_size=convert_es->nb_space;
  e->data=malloc(sizeof(char *)*convert_es->nb_data);
  e->data_size=convert_es->nb_data;
  
  struct evtstructelem *ese=convert_es->first;
  char *field;
  while(ese!=NULL) {
    field=strdup(get_event_field(source,ese->name));
    if (ese->domain==TIME) {
      if (ese->index>e->time_size-1)
        printf("warning : ese named %s has index %d incompatible with time size %d\n",ese->name,ese->index,e->time_size);
      e->time[ese->index]=field;
    }
    if (ese->domain==SPACE) {
      if (ese->index>e->space_size-1)
        printf("warning : ese named %s has index %d incompatible with space size %d\n",ese->name,ese->index,e->space_size);
      e->space[ese->index]=field;
    }
    if (ese->domain==DATA) {
      if (ese->index>e->data_size-1)
        printf("warning : ese named %s has index %d incompatible with data size %d\n",ese->name,ese->index,e->data_size);
      e->data[ese->index]=field;
    }
    ese=ese->next;
  }
  add_event(b,e);
}

int check_block(struct block *bl) {
  int i;
  int res;
  struct event *e;
  if (bl==NULL) {
    printf("error : block is null\n");
    return -1;
  }
  if (bl->es==NULL) {
    printf("error : block has no schema\n");
    return -1;
  }
  if (bl->es->nb_props!=bl->props_size) {
    printf("error : block nb props does not equal props size\n");
    return -1;
  }
  for(i=0;i<bl->es->nb_props;i++) {
    if (bl->props[i]==NULL) {
       printf("error : block field props[%d] is NULL\n",i);
      return -1;
    }
  }
  e=bl->events;
  while(e!=NULL) {
    res=check_event(e);
    if (res==-1)
      return -1;
    e=e->next;
  }
  return 0;
}

int check_event(struct event *e) {
  int i;
  if (e==NULL) {
    printf("error : event is null\n");
    return -1;
  }
  if (e->block==NULL) {
    printf("error : event block is null\n");
    return -1;
  }
  if (e->block->es==NULL) {
    printf("error : event block schema is null\n");
    return -1;
  }

  if (e->block->es->nb_time!=e->time_size) {
    printf("error : event nb time does not correspond to time size\n");
    return -1;
  }
  for(i=0;i<e->block->es->nb_time;i++) {
    if (e->time[i]==NULL) {
      printf("error : event field time[%d] is NULL\n",i);
      return -1;
    }
  }

  if (e->block->es->nb_space!=e->space_size) {
    printf("error : event nb space does not correspond to space size\n");
    return -1;
  }
  for(i=0;i<e->block->es->nb_space;i++) {
    if (e->space[i]==NULL) {
       printf("error : event field space[%d] is NULL\n",i);
      return -1;
    }
  }

  if (e->block->es->nb_data!=e->data_size) {
    printf("error : event nb data does not correspond to data size\n");
    return -1;
  }
  for(i=0;i<e->block->es->nb_data;i++) {
    if (e->data[i]==NULL) {
      printf("error : event field data[%d] is NULL\n",i);
      return -1;
    }
  }

  return 0;
  
}



char exists_prop(struct evtstruct *es,char *name) {
  struct evtstructelem *ese;
  ese=es->first;
  while(ese!=NULL) {
    if ((!strcmp(ese->name,name)) && (ese->domain==PROPS))
      return 1;
    ese=ese->next;
  }
  return 0;
}
