/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include <pyrame.h>
int semtimedop(int semid, struct sembuf *sops, unsigned nsops,struct timespec *timeout);

//*************************** SEMAPHORES *********************************

int open_semaphore(key_t key, unsigned char create,int setval) {
  int result;
  int semaphore;
  if (create) {
    semaphore=semget(key,1,IPC_CREAT|IPC_EXCL|0666);
    if ((semaphore==-1) && (errno==EEXIST)) {
	semaphore=semget(key,1,0);
	semctl(semaphore,0,IPC_RMID);
	semaphore=semget(key,1,IPC_CREAT|IPC_EXCL|0666);
      }
  } else {
    semaphore=semget(key,1,0);
  }
  if (semaphore==-1) {
    printf("cant open semaphore with key %d\n",key);
    perror("semget");
    exit(1);
  }
  if (create) {
    result=semctl(semaphore,0,SETVAL,setval);
    if (result==-1) {
      perror("semctl");
      exit(1);
    }
  }
  return semaphore;
}

void destroy_semaphore(int semaphore) {
  semctl(semaphore,0,IPC_RMID);
}

//acquire the semaphore
//return -1 if an error occurs during the semop
int P(int semid) {
  struct sembuf op;
  int result;
  op.sem_num=0;
  op.sem_op=-1;
  op.sem_flg=0;
  result=semop(semid,&op,1);
  if (result==-1) {
    perror("semop");
    return -1;
  }
  return 0;
}

struct timespec *ms2ts(unsigned long ms)
{
  struct timespec *result=malloc(sizeof(struct timespec));
  result->tv_sec=ms/1000;
  result->tv_nsec=(ms%1000)*1000000;
  return result;
}

//acquire the semaphore
//return -1 if an error occurs during the semop
int timeP(int semid,unsigned long ms) {
  struct sembuf op;
  int result;
  struct timespec *timeout=ms2ts(ms);
  op.sem_num=0;
  op.sem_op=-1;
  op.sem_flg=0;
  result=semtimedop(semid,&op,1,timeout);
  free(timeout);
  if (result==-1) {
    //perror("semop");
    return -1;
  }
  return 0;
}

//free the semaphore
//return -1 if an error occurs during the semop
int V(int semid) {
  struct sembuf op;
  int result;
  op.sem_num=0;
  op.sem_op=1;
  op.sem_flg=0;
  result=semop(semid,&op,1);
  if (result==-1) {
    perror("semop");
    return -1;
  }
  return 0;
}

int semread(int semid) {
  int res;
  res=semctl(semid,0,GETVAL,0);
  if (res==-1) {
    perror("semctl");
    return -1;
  } else {
    return res;
  }
}


//*********************** SHARED MEMORY ******************************** 

int open_shmem(key_t key,int create) {
  int shmid;
  if (create) {
    shmid=shmget(key,SHARED_BUFFER_SIZE,IPC_CREAT|0666);
  } else {
    shmid=shmget(key,SHARED_BUFFER_SIZE,0);
  }
  if (shmid==-1) {
    printf("cant open shmem with key %d\n",key);
    perror("shmget");
    exit(1);
  }
  return shmid;
}

void * get_shmem_pointer(int shmid) {
  return shmat(shmid,(void *)0,0);
}

void destroy_shmem(int shmid,void *shdata) {
  shmdt(shdata);
  shmctl(shmid,IPC_RMID,NULL);
}


//************************* SHBUFFER **************************************

//open an shbuffer
//create value is 1 for create the shbuffer (data sender) 0 for using existing one (data receiver)
struct shbuffer * open_shbuffer(int port_id,int create,int timeout) {
  struct shbuffer *result=malloc(sizeof(struct shbuffer));
  memset(result,0,sizeof(struct shbuffer));
  result->full_semid=open_semaphore((key_t)(20000+port_id*10+1),create,0);
  result->empty_semid=open_semaphore((key_t)(20000+port_id*10+2),create,1);
  result->mutex_semid=open_semaphore((key_t)(20000+port_id*10+3),create,0);
  result->ready_semid=open_semaphore((key_t)(20000+port_id*10+4),create,0);
  result->shmid=open_shmem((key_t)(20000+port_id*10),create);
  result->portid=port_id;
  result->real_shared_buffer=get_shmem_pointer(result->shmid);
  result->shdata=result->real_shared_buffer+2;
  result->flushed=1;
  result->tmp_size=0;
  result->tmp=malloc(MAX_DATA_SIZE*sizeof(char));
  result->data_size=0;
  result->timeout=timeout;
  gettimeofday(&result->last_flush,NULL);
  pthread_mutex_init(&result->data_mutex,NULL);
  if (create) {
    if (pthread_create(&(result->t_monitor),NULL,monitor_shbuffer,result)<0) {
    perror("pthread_create");
    exit(1);
  }
  pthread_detach(result->t_monitor);
  }
  return result;
}

void flush_shbuffer(struct shbuffer *shbuf) {
  shbuf->data_size=0;
  shbuf->tmp_size=0;
}

void close_shbuffer(struct shbuffer *shbuf) {
  destroy_shmem(shbuf->shmid,shbuf->shdata);
  destroy_semaphore(shbuf->full_semid);
  destroy_semaphore(shbuf->empty_semid);
  destroy_semaphore(shbuf->mutex_semid);
  destroy_semaphore(shbuf->ready_semid);
  free(shbuf->tmp);
  pthread_cancel(shbuf->t_monitor);
  free(shbuf);
}


int isready_shbuffer(struct shbuffer *shbuf) {
  return semread(shbuf->ready_semid);
}

void ready_shbuffer(struct shbuffer *shbuf) {
  int result;
  //printf("READY : ready_shbuffer %d\n",shbuf->portid);
  result=semread(shbuf->ready_semid);
  if (result==0)
    V(shbuf->ready_semid);
  printf("shbuffer %d is ready\n",shbuf->portid);
}

void unready_shbuffer(struct shbuffer *shbuf) {
  int result;
  result=semread(shbuf->ready_semid);
  if (result==1)
    P(shbuf->ready_semid);
  semctl(shbuf->full_semid,0,SETVAL,0);
  semctl(shbuf->empty_semid,0,SETVAL,0);
  semctl(shbuf->mutex_semid,0,SETVAL,0);
}

//read the content of a shbuffer
//return the size of the read and -1 in case of semop error
int read_shbuffer(struct shbuffer *shbuf,char *buffer) {
  unsigned short rsize;
  int res;
  //printf("read_shbuffer\n");
  
  //printf("wait for full semid\n");
  res=timeP(shbuf->full_semid,3000);
  if (res==-1) {
    //printf("read_shbuffer aborted because of full_semid semop error\n");
    return -1;
  }
  //printf("wait for mutex semid\n");
  timeP(shbuf->mutex_semid,3000);
  if (res==-1) {
    //printf("read_shbuffer aborted because of mutex_semid semop error\n");
    return -1;
  }
  rsize=*(unsigned short *)shbuf->real_shared_buffer;
  //printf("read size=%d\n",rsize);
  memcpy(buffer,shbuf->shdata,rsize);
  //printf("releasing mutex semid\n");
  V(shbuf->mutex_semid);
  //printf("releasing empty semid\n");
  V(shbuf->empty_semid);
  return rsize;
}


void publish_shbuffer(struct shbuffer *shbuf) {

  int res;
  //signalize the client to read data
  //printf("releasing mutex semid\n");
  V(shbuf->mutex_semid);
  //printf("releasing full semid\n");
  V(shbuf->full_semid);
  
  //wait until data is read
  //printf("wait for empty semid\n");
  res=timeP(shbuf->empty_semid,3000);
  if (res==-1) {
    printf("warning : write_shbuffer aborted on empty_semid-> unready\n");
    unready_shbuffer(shbuf);
    return;
  }
  //printf("wait for mutex semid\n");
  res=timeP(shbuf->mutex_semid,3000);
  if (res==-1) {
    printf("warning write_shbuffer aborted on mutex_semid-> unready\n");
    unready_shbuffer(shbuf);
    return;
  }
  
  //if some data remains from last send, copy it at the beginning of buffer
  shbuf->data_size=0;
  if (shbuf->tmp_size) {
    memcpy(shbuf->shdata,shbuf->tmp,shbuf->tmp_size);
    shbuf->data_size=shbuf->tmp_size;
    shbuf->tmp_size=0;
  }
  shbuf->flushed=0;
}


void check_timeout(struct shbuffer *shbuf) {
  struct timeval now;
  int delta_time;
  
  if (!isready_shbuffer(shbuf))
    return;
  
  //compute the delta time since last flush
  gettimeofday(&now,NULL);
  delta_time=(now.tv_sec*1000000+now.tv_usec)-(shbuf->last_flush.tv_sec*1000000+shbuf->last_flush.tv_usec);
  
  //if the shmem is full or timeout is over -> release it
  //printf("comparing delta_time=%d and timeout=%d\n",delta_time,shbuf->timeout);
  if ((shbuf->data_size==MAX_DATA_SIZE) || (delta_time>shbuf->timeout && shbuf->data_size!=0)) {
    //printf("flushing\n");
    if (delta_time>shbuf->timeout) {
      //printf("flushing on timeout size=%d\n",shbuf->data_size);
    }
    shbuf->flushed=1;
    shbuf->last_flush.tv_sec=now.tv_sec;
    shbuf->last_flush.tv_usec=now.tv_usec;
    *(unsigned short *)shbuf->real_shared_buffer=(unsigned short)shbuf->data_size;
    //printf("flush time : %d:%d\n",shbuf->last_flush.tv_sec,shbuf->last_flush.tv_usec);
    
    publish_shbuffer(shbuf);
  }
}

void *monitor_shbuffer(void *vdata) {
  struct shbuffer *shbuf=(struct shbuffer *)vdata;
  while(1) {

    //printf("monitor waiting for mutex\n");
    //acquire data mutex
    pthread_mutex_lock(&shbuf->data_mutex);
    //printf("monitor acquired mutex\n");

    check_timeout(shbuf);

    //release the data mutex
    pthread_mutex_unlock(&shbuf->data_mutex);
    //printf("monitor released mutex\n");

    usleep(shbuf->timeout);
  }
  return NULL;
}

//write to a shbuffer
//return the written size or -1 in case of semop error
int write_shbuffer(struct shbuffer *shbuf,char *buffer,int size) {
  int fill_size;

  //printf("write_shbuffer of size %d\n",size);

  //acquire data mutex
  //printf("write waiting for mutex\n");
  pthread_mutex_lock(&shbuf->data_mutex);
  //printf("write acquired mutex\n");
  
  if ((shbuf->data_size+size)<=MAX_DATA_SIZE) {
    //the data is too short to fill the entire shdata so just copy and wait another packet
    memcpy(shbuf->shdata+shbuf->data_size,buffer,size);
    shbuf->data_size+=size;
  }  else {
    //the packet size exceed the shdata size so copy the overrun in tmp
    fill_size=MAX_DATA_SIZE-shbuf->data_size;
    memcpy(shbuf->shdata+shbuf->data_size,buffer,fill_size);
    shbuf->tmp_size=size-fill_size;
    memcpy(shbuf->tmp,buffer+fill_size,shbuf->tmp_size);
    shbuf->data_size=MAX_DATA_SIZE;
  }

  //check if buffer should be released
  check_timeout(shbuf);

  //release the data mutex
  //printf("write releasing mutex\n");
  pthread_mutex_unlock(&shbuf->data_mutex);
  //printf("write released mutex\n");


  return size;
}
