/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include <pyrame.h>

//initialize the stats system
struct stats_ws *init_stats(char *varmod_ip,int periodicity) {
  struct stats_ws *ws=(struct stats_ws *)malloc(sizeof(struct stats_ws));
  memset(ws,0,sizeof(struct stats_ws));

  ws->varmod_socket=-1;
  if (varmod_ip==NULL) {
    printf("Warning: No varmod available -> stats varmod update disabled\n");
    ws->vip=NULL;
  } else {
    ws->vip=strdup(varmod_ip);
    ws->vport=get_port("VARMOD_PORT");
  }
  
  //stats names and values
  ws->stats=(struct pyr_stat **)malloc(MALLOC_STEP*sizeof(struct pyr_stat*));
  ws->nb_stats=0;
  ws->malloced=MALLOC_STEP;
  ws->block_update=0;
  ws->socket_pb=0;

  //stats mutex
  ws->stat_mutex=malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(ws->stat_mutex,NULL);
  
  //launch the thread
  ws->periodicity=periodicity;
  ws->t_stats=(pthread_t*)malloc(sizeof(pthread_t));
  if (pthread_create(ws->t_stats,NULL,thread_stats,ws)<0) {
    perror("pthread_create");
    exit(1);
  }
  pthread_detach(*ws->t_stats);

  return ws;
}

//allocate a new statistic structure CAUTION : no mutex taken
struct pyr_stat *alloc_next_stat(struct stats_ws *ws) {
  struct pyr_stat *res;
  if (ws->nb_stats==ws->malloced) {
    ws->malloced+=MALLOC_STEP;
    ws->stats=(struct pyr_stat **)realloc(ws->stats,ws->malloced*sizeof(struct pyr_stat*));
  }
  ws->stats[ws->nb_stats]=(struct pyr_stat *)malloc(sizeof(struct pyr_stat));
  res=ws->stats[ws->nb_stats];
  ws->nb_stats++;
  return res;
}

void add_acc(struct stats_ws *ws,char *name,double init_value) {
  struct pyr_stat *new_stat;
  pthread_mutex_lock(ws->stat_mutex);
  new_stat=alloc_next_stat(ws);
  new_stat->name=strdup(name);
  new_stat->type=TYPE_ACC;
  new_stat->value=init_value;
  new_stat->varvalue=init_value;
  new_stat->initialized=0;
  pthread_mutex_unlock(ws->stat_mutex);
}

void add_inst(struct stats_ws *ws,char *name,double init_value) {
  struct pyr_stat *new_stat;
  pthread_mutex_lock(ws->stat_mutex);
  new_stat=alloc_next_stat(ws);
  new_stat->name=strdup(name);
  new_stat->type=TYPE_INST;
  new_stat->value=init_value;
  new_stat->varvalue=init_value;
  new_stat->initialized=0;
  pthread_mutex_unlock(ws->stat_mutex);
}

void add_rate(struct stats_ws *ws,char *name,char *ref) {
  struct pyr_stat *new_stat;
  pthread_mutex_lock(ws->stat_mutex);
  new_stat=alloc_next_stat(ws);
  new_stat->name=strdup(name);
  new_stat->type=TYPE_RATE;
  new_stat->ref=strdup(ref);
  new_stat->value=0;
  new_stat->varvalue=0;
  gettimeofday(&(new_stat->last_time),NULL);
  new_stat->initialized=0;
  pthread_mutex_unlock(ws->stat_mutex);
}

void add_sw(struct stats_ws *ws,char *name,char *ref,int win_size) {
  struct pyr_stat *new_stat;
  int i;
  pthread_mutex_lock(ws->stat_mutex);
  new_stat=alloc_next_stat(ws);
  new_stat->name=strdup(name);
  new_stat->type=TYPE_SW;
  new_stat->ref=strdup(ref);
  new_stat->value=0;
  new_stat->varvalue=0;
  new_stat->win_size=win_size;
  new_stat->mem=malloc(sizeof(double)*win_size);
  for(i=0;i<new_stat->win_size;i++) {
    new_stat->mem[i]=0;
  }
  new_stat->initialized=0;
  pthread_mutex_unlock(ws->stat_mutex);
}

void add_pc(struct stats_ws *ws,char *name,char *ref,char *refsum) {
  struct pyr_stat *new_stat;
  pthread_mutex_lock(ws->stat_mutex);
  new_stat=alloc_next_stat(ws);
  new_stat->name=strdup(name);
  new_stat->type=TYPE_PC;
  new_stat->ref=strdup(ref);
  new_stat->refsum=strdup(refsum);
  //printf("for pc named %s numerator=%s denominator=%s\n",name,new_stat->ref,new_stat->refsum);
  new_stat->value=0;
  new_stat->varvalue=0;
  new_stat->initialized=0;
  pthread_mutex_unlock(ws->stat_mutex);
}

//get the value for a statistic CAUTION : no mutex taken
double get_value(struct stats_ws *ws,char *name) {
  int i;
  for(i=0;i<ws->nb_stats;i++) {
    if (!strcmp(ws->stats[i]->name,name)) {
      return ws->stats[i]->value;
    }
  }
  return 0;
}

//get the value for a statistic CAUTION : no mutex taken
double get_varvalue(struct stats_ws *ws,char *name) {
  int i;
  for(i=0;i<ws->nb_stats;i++) {
    if (!strcmp(ws->stats[i]->name,name)) {
      return ws->stats[i]->varvalue;
    }
  }
  return 0;
}

//get a statistic by name CAUTION : no mutex taken
struct pyr_stat *get_stat(struct stats_ws *ws,char *name) {
  int i;
  for(i=0;i<ws->nb_stats;i++) {
    if (!strcmp(ws->stats[i]->name,name)) {
      return ws->stats[i];
    }
  }
  return NULL;
}


void free_stat(struct pyr_stat *s) {
  if (s->type==TYPE_RATE) {
    free(s->ref);
  }
  free(s->name);
  free(s);
}

//initialize the stats system
void deinit_stats(struct stats_ws *ws) {
  int i;
  pthread_mutex_lock(ws->stat_mutex);
  pthread_cancel(*ws->t_stats);
  if (ws->varmod_socket!=-1)
    close(ws->varmod_socket);
  ws->varmod_socket=-1;
  free(ws->vip);
  for (i=0;i<ws->nb_stats;i++) 
    free_stat(ws->stats[i]);
  free(ws->stats);
  free(ws->stat_mutex);
  free(ws);
}


//allow to update any stat value
void set_stats(struct stats_ws *ws,char *name,int op,double value) {
  struct pyr_stat *cur_stat;
  pthread_mutex_lock(ws->stat_mutex);
  //printf("set stats %s op=%d value=%f\n",name,op,value);
  cur_stat=get_stat(ws,name);
  if (cur_stat!=NULL) {
    //printf("old value=%f\n",cur_stat->value);
    switch(op) {
    case SOP_PLUS:
      cur_stat->value+=value;
      cur_stat->varvalue+=value;
      break;
    case SOP_MIN:
      cur_stat->value-=value;
      cur_stat->varvalue-=value;
      break;
    case SOP_SET:
      cur_stat->value=value;
      break;
    }
  }
  //printf("new value=%f\n",cur_stat->value);
  pthread_mutex_unlock(ws->stat_mutex);
  return;
}


//return all the current stats in a string
char * get_stats(struct stats_ws *ws) {
  struct incr_string *ires=new_is();
  int i;
  pthread_mutex_lock(ws->stat_mutex);
  for (i=0;i<ws->nb_stats;i++) {
    fsuffix_is(ires,"%s=%f ",ws->stats[i]->name,ws->stats[i]->value);
  }
  pthread_mutex_unlock(ws->stat_mutex);
  return reduce_is(ires);
}

//dump the stats in xml in the given file
void dump_stats_to_log(struct stats_ws *ws,int log_file) {
  char *res=malloc(4096*sizeof(char));
  struct cmd_result *result;
  int i;
  int wres;

  //sanity check
  if (log_file==-1)
    return;
  if (ws->varmod_socket == -1)
      return;

  //block auto update
  ws->block_update=1;
  //printf("waiting for update to finish\n");
  while(ws->in_update==1) {
    usleep(500000);
  }
  //printf("beginning of dump\n");
  
  //compute and update prior to dump
  compute_stats(ws);
  update_stats(ws);
  
  //dump values in xml
  pthread_mutex_lock(ws->stat_mutex);
  for (i=0;i<ws->nb_stats;i++) {
    result=execcmd(ws->varmod_socket,"getvar_varmod","varmod_id","default",ws->stats[i]->name,"end");
    if (result->retcode==0) {
      printf("error on getting stat : %s\n",result->str);
      ws->socket_pb=1;
    }
    sprintf(res,"    <param name=\"%s\">%s</param>\n",ws->stats[i]->name,result->str);
    wres=write(log_file,res,strlen(res));
    if (wres==0)
      printf("nothing written\n");
    if (wres==-1) {
      perror("write");
    }
    free_cmd_result(result);
  }
  free(res);
  pthread_mutex_unlock(ws->stat_mutex);
  ws->block_update=0;
  //printf("end of dump\n");
}

//zero all stats
void zero_stats(struct stats_ws *ws) {
  int i;
  struct cmd_result *result;
  if (ws->varmod_socket == -1)
      return;
  pthread_mutex_lock(ws->stat_mutex);
  for (i=0;i<ws->nb_stats;i++) {
    ws->stats[i]->value=0;
    ws->stats[i]->varvalue=0;
    result=execcmd(ws->varmod_socket,"setvar_varmod","varmod_id","default",ws->stats[i]->name,"0","end");
    if (result->retcode==0) {
      printf("error on setting stat : %s\n",result->str);
      ws->socket_pb=1;
    }
    free_cmd_result(result);
  }
  pthread_mutex_unlock(ws->stat_mutex);
}

void compute_stats(struct stats_ws *ws) {
  double ref_value;
  double last_time;
  double cur_time;
  double interval;
  double numerator,denominator;
  double sum;
  struct pyr_stat *cs; //current stat for code compacity
  int i,j;
  pthread_mutex_lock(ws->stat_mutex);
  for (i=0;i<ws->nb_stats;i++) {
    //printf("evaluating stats %d of name %s\n",i,ws->stats[i]->name);
    cs=ws->stats[i];
    if (cs->type==TYPE_ACC) {
    }

    
    if (cs->type==TYPE_INST) {
    }

    
    if (cs->type==TYPE_RATE) {
      ref_value=get_value(ws,cs->ref);
      last_time=cs->last_time.tv_usec+cs->last_time.tv_sec*1000000;
      gettimeofday(&(cs->last_time),NULL);
      cur_time=cs->last_time.tv_usec+cs->last_time.tv_sec*1000000;
      interval=(cur_time-last_time)/1000000.0;
      if (interval==0.0) {
        cs->value=0;
      } else {
        cs->value=ref_value/interval;
      }
    }
    
    
    if (cs->type==TYPE_SW) {
      //shifting all value on the left
      for(j=cs->win_size-1;j>0;j--) {
        cs->mem[j]=cs->mem[j-1];
      }   
      //adding the new value at the beginning
      cs->mem[0]=get_value(ws,cs->ref);
      //calculate the mean on the window
      sum=0;
      //printf("window=");
      for(j=0;j<cs->win_size;j++) {
        sum+=cs->mem[j];
        //printf("%f ",cs->mem[j]);
      }
      cs->value=sum/cs->win_size;
      //printf("-> moy=%f\n",cs->value);
    }
    
    if (cs->type==TYPE_PC) {
      numerator=get_value(ws,cs->ref);
      denominator=get_value(ws,cs->refsum);
      if (denominator!=0) {
        cs->value=(float)numerator/(float)denominator*100.0;
      }
      //printf("pc=%f/%f=%f\n",numerator,denominator,cs->value);
    }
  }
  pthread_mutex_unlock(ws->stat_mutex);
}

//compute all the stats and send it to varmod 
void update_stats(struct stats_ws *ws) {
  int i;
  char value[1024];
  struct cmd_result *result=NULL;
  struct pyr_stat *cs; //current stat for code compacity
  
  pthread_mutex_lock(ws->stat_mutex);
  if (ws->varmod_socket != -1) {
    for (i=0;i<ws->nb_stats;i++) {
      cs=ws->stats[i]; 

      //communicate with varmod
      if (cs->type==TYPE_ACC) {
        if (cs->initialized==0) {
          sprintf(value,"%f",cs->varvalue);
          result=execcmd(ws->varmod_socket,"setvar_varmod","varmod_id","default",cs->name,value,"end");
        } else {
          if (cs->varvalue==0) {
            result=NULL;
          } else {
            sprintf(value,"%f",cs->varvalue);
            result=execcmd(ws->varmod_socket,"flopvar_varmod","varmod_id","default",cs->name,"+",value,"end");
          }
        }
      }
      if (cs->type==TYPE_INST || cs->type==TYPE_RATE || cs->type==TYPE_SW || cs->type==TYPE_PC) {
        if (cs->value==cs->varvalue) {
          result=NULL;
        } else {
          sprintf(value,"%f",cs->value);
          result=execcmd(ws->varmod_socket,"setvar_varmod","varmod_id","default",cs->name,value,"end");
        }
      }

      //post treatment
      if (result!=NULL) {
        if (result->retcode==0) {
          //printf("error on setting stat : %s\n",result->str);
          ws->socket_pb=1;
        } else {
          //printf("%s updated with value %s\n",cs->name,value);
          if (cs->type==TYPE_ACC) {
            cs->varvalue=0; //dont accumulate twice
          } else {
            //store the varmod value to avoid next send if =
            cs->varvalue=cs->value; 
          }
          if (cs->initialized==0) {
            cs->initialized=1;
            //printf("%s initialized\n",cs->name);
          }
        }
        free_cmd_result(result);
      }
      
    }//loop on stats
  }
  pthread_mutex_unlock(ws->stat_mutex);
}

void check_socket(struct stats_ws * ws) {
  int i;
  struct pyr_stat *cs;
  if (ws->socket_pb==1) {
    close(ws->varmod_socket);
    ws->varmod_socket=-1;
    ws->socket_pb=0;
    pthread_mutex_lock(ws->stat_mutex);
    for (i=0;i<ws->nb_stats;i++) {
      cs=ws->stats[i]; 
      if (cs->type==TYPE_ACC) {
        cs->varvalue=cs->value;
      } else {
        cs->varvalue=0;
      }
    }
    pthread_mutex_unlock(ws->stat_mutex);
  }
  if (ws->varmod_socket==-1) {
    ws->varmod_socket=open_socket(ws->vip,ws->vport);
  }
}


//update the stats every second
void *thread_stats(void *args) {
  struct stats_ws *ws=(struct stats_ws *)args;
  while(1) {
    if (!ws->block_update) {
      ws->in_update=1;
      check_socket(ws);
      compute_stats(ws);
      update_stats(ws);
      usleep(ws->periodicity);
      ws->in_update=0;
    }
  }
}

