#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy, LLR
#

import math

def split_bin(n,size,complement=False):
    "translate an integer in binary string"
    if not size:
        return ""
    if complement:
        n=~n&(2**size-1)
    return ("{0:0%sb}"%size).format(n)[-size:]

def is_hexa(s):
    "test if a string represent an hexadecimal number"
    try:
        int("0x%s" % s,16)
        return True
    except ValueError:
        return False

def load_file(filename,bs_length,inverse=False):
    "Load a bitstream from a file. If the file content starts with '0x' or '0b ' (note the whitespace), it is considered to be hexadecimal or binary respectively. Otherwise it is considered a decimal integer."
    try:
        fd=open(filename,"r")
    except:
        return 0,"can't open file %s"%(filename)
    try:
        bitstream=fd.read()
    finally:
        fd.close()
    return load_str(bitstream,bs_length,inverse)

def load_str(bs,bs_length,inverse=False):
    "Load a bitstream from string. If the string starts with '0x' or '0b' (case sensitive), it is considered to be hexadecimal or binary respectively. Otherwise it is considered a decimal integer."
    bs=bs.strip()
    if bs[:2]=="0b":
        try:
            int(bs[2:],2)
        except Exception as e:
            return 0,"invalid binary bitstream: %s"%(str(e))
        bs_b=bs[2:]
    elif bs[:2]=="0x":
        bs=bs[2:]
        try:
            bs_b=split_bin(int(bs,16),4*len(bs))
        except Exception as e:
            return 0,"invalid hexadecimal bitstream: %s"%(str(e))
    else:
        try:
            bs=int(bs)
            if bs!=0:
                bs_b=split_bin(bs,int(math.ceil(math.log(bs,2))))
            else:
                bs_b="0"
        except Exception as e:
            return 0,"invalid decimal bitstream: %s"%(str(e))
    len_read=len(bs_b)
    if len_read>bs_length:
        bs_b=bs_b[:bs_length]
        print("\n\nWarning: trimming bitstream to %d. It was %d\n\n"%(bs_length,len_read))
    if len_read<bs_length:
        return 0,"Invalid bitstream length %d. Should be %d"%(len_read,bs_length)
    if inverse:
        bs_b=bs_b[::-1]
    return 1,bs_b

def apply_mask(roc,pos,src):
    "apply a pattern in a bitstream at fixed position"
    if src!="undef":
        lpart=roc["bitstream_bin"][0:pos]
        rpart=roc["bitstream_bin"][pos+len(src):len(roc["bitstream_bin"])]
        roc["bitstream_bin"]=lpart+src+rpart
    return 1,"ok"

def apply_inv_mask(roc,pos,src):
    "the same but with inverse order"
    #reverse the pattern
    apat=src[::-1]
    return apply_mask(roc,pos,apat)

def bin2gray(num):
    "convert to gray"
    return (num>>1)^num

def bin2hexa(binstr,msb_padding=True):
    "convert a binary string in hexadecimal, padding with zeroes on MSB if necessary if msb_padding is True or absent and padding on LSB otherwise"
    if binstr.startswith("0b"):
        binstr=binstr[2:]
    if not msb_padding:
        binstr=binstr+"0"*(int(math.ceil(len(binstr)/4.0))*4-len(binstr))
    return ("{0:0%dx}"%(int(math.ceil(len(binstr)/4.0)))).format(int(binstr,2))

def parse_channel_list(channel_list):
    "Parse comma-separated channel list with dash-defined ranges accepted. e.g.: 1,3,7-12,5"
    if channel_list=="undef" or channel_list=="":
        return []
    if channel_list=="all":
        channel_list="0-%d"%(nb_chans-1)
    channel_list=channel_list.split(",")
    channels=[]
    for channel in channel_list:
        if len(channel.split("-")) > 1:
            channels += range(int(channel.split("-")[0]),int(channel.split("-")[1])+1)
        else:
            channels.append(int(channel))
    if max(channels)>=nb_chans:
        raise Exception("channel(s) over maximum number of channels (%d)"%(nb_chans))
    return list(set(channels))

def extract_bit_chans(roc,start,every,check,inverse=True):
    "Returns list of channels with bit=*check* in bitstream of *roc*. *start* is the bit position of channel 0 if *inverse* is False or the last channel, otherwise. Successive channels are separated by *every* bits."
    if check not in [0,1]:
        raise Exception("check bit must be 0 or 1")
    check=str(check)
    bs=roc["bitstream_bin"][start:start+(nb_chans-1)*every+1]
    chans=[]
    for chan in range(nb_chans):
        if inverse:
            pos=(nb_chans-1-chan)*every
        else:
            pos=chan*every
        if bs[pos:pos+1]==check:
            chans.append(chan)
    return ",".join(map(str,chans))

def not_bit(bit):
    if bit not in ["1","0"]:
        raise Exception("not_bit: bit must be 0 or 1")
    if bit=="1":
        return "0"
    else:
        return "1"

def on_off_bit(roc,pos,value,inv=False):
    value=str(value).upper().strip()
    if value=="ON" or value=="1" or value=="TRUE":
        value="1"
    elif value=="OFF" or value=="0" or value=="FALSE":
        value="0"
    elif value=="UNDEF":
        return 1,"ok"
    elif value!="UNDEF":
        return 0,"invalid value: %s"%(value)
    if inv:
        return apply_mask(roc,pos,not_bit(value))
    return apply_mask(roc,pos,value)

def update_bitstream(roc):
    roc["bitstream"]="0x"+bin2hexa(roc["bitstream_bin"],msb_padding=False)

def apply_to_chans(roc_pool,roc_id,chans,extract_func,set_func,xml_param,*params):
    "generic application of a function to multiple channels of a roc"
    try:
        roc=roc_pool.get(roc_id)
    except Exception as e:
        return 0,str(e)
    if chans=="undef":
        return 1,"ok"
    #apply the bitstream modification
    for c in parse_channel_list(chans):
        func_name=set_func.__name__
        if func_name=="on_off_bit" and xml_param!="undef":
            func_name=xml_param
        retcode,res=set_func(roc,c,*params)
        if retcode==0:
            return 0,"can't %s on roc %s channel %s: %s"%(func_name,roc_id,c,res)
    if xml_param!="undef":
        roc[xml_param]=extract_func(roc)
    #update the bitstream in cmod
    update_bitstream(roc)
    return 1,"ok"

def apply_to_roc(roc_pool,roc_id,extract_func,set_func,xml_param,*params):
    "generic application of a function to roc"
    try:
        roc=roc_pool.get(roc_id)
    except Exception as e:
        return 0,str(e)
    old_bs=roc["bitstream_bin"]
    retcode,res=set_func(roc,*params)
    if retcode==0:
        return 0,"can't set %s on roc %s: %s"%(xml_param,roc_id,res)
    if xml_param!="undef":
        roc[xml_param]=extract_func(roc)
    #if bitstream hasn't changed, no need to update bitstream in cmod
    if old_bs==roc["bitstream_bin"]:
        return 1,"ok"
    #update the bitstream in cmod
    update_bitstream(roc)
    return 1,"ok"


def get_param_roc(roc_pool,roc_id,param_name):
    "generic parameter extraction of a roc"
    try:
        roc=roc_pool.get(roc_id)
    except Exception as e:
        return 0,str(e) 
    if param_name not in roc:
        return 0,"the parameter does not exists in this roc"
    return 1,roc[param_name]

