#!/usr/bin/env python2
# -*- coding: utf-8 -*-
# 
# Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
# This file is part of Pyrame.
# 
# Pyrame is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# Pyrame is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrame.  If not, see <http://www.gnu.org/licenses/>

import time
import pools,conf_strings,scpi

ag_81160_pool=pools.pool()

# Functions

class ag_81160_class(scpi.scpi):
    channels=["1","2"]

    def __init__(self):
        super(ag_81160_class,self).__init__("ag_81160")

    def reset(self):
        retcode,res=super(ag_81160_class,self).reset()
        if retcode==0:
            return 0,res
        command=r"OUTP1:TRIG:ROUT NONE\n"
        command+="OUTP1:STR:ROUT NONE"
        return self.simple_command(ag_81160_id,command)

    def set_function(self,ag_81160_id,function):
        if function not in ["sine","square","ramp","pulse"]:
            return 0,"invalid function"
        value={"sine":"SIN","square":"SQU","ramp":"RAMP","pulse":"PULS"}.get(function)
        command="FUNC{channel} %s"%(value)
        return self.simple_command(ag_81160_id,command)

    def set_frequency(self,ag_81160_id,frequency):
        if frequency=="undef":
            return 1,"ok"
        command="FREQ{channel} %.12e"%(float(frequency))
        return self.simple_command(ag_81160_id,command)

    def get_frequency(self,ag_81160_id):
        command="FREQ{channel}?"
        return self.simple_query(ag_81160_id,command)

    def set_high_level(self,ag_81160_id,high_level):
        if high_level=="undef":
            return 1,"ok"
        command="VOLT{channel}:HIGH %.5e"%(float(high_level))
        return self.simple_command(ag_81160_id,command)

    def set_low_level(self,ag_81160_id,low_level):
        if low_level=="undef":
            return 1,"ok"
        command="VOLT{channel}:LOW %.5e"%(float(low_level))
        return self.simple_command(ag_81160_id,command)

    def set_duty_cycle(self,ag_81160_id,duty_cycle):
        if duty_cycle=="undef":
            return 1,"ok"
        command="FUNC{channel}:SQU:DCYC %.9e"%(float(duty_cycle))
        return self.simple_command(ag_81160_id,command)

    def set_pulse_width(self,ag_81160_id,pulse_width):
        if pulse_width=="undef":
            return 1,"ok"
        command="FUNC{channel}:PULS:WIDT %.4e"%(float(pulse_width))
        return self.simple_command(ag_81160_id,command)

    def set_edges(self,ag_81160_id,rising_edge,falling_edge):
        # Verify edges
        if rising_edge.lower() not in ["min","max","undef"]:
            value_r="%.9f"%(float(rising_edge))
        else:
            value_r=rising_edge
        if falling_edge.lower() not in ["min","max","undef"]:
            value_f="%.9f"%(float(falling_edge))
        else:
            value_f=falling_edge
        # Send the command
        command=""
        if value_r!="undef":
            command+="SOUR{channel}:FUNC:PULS:TRAN:LEAD %s"%(value_r)
        if "undef" not in [value_r,value_f]:
            command+=r"\n"
        if value_f!="undef":
            command+="SOUR{channel}:FUNC:PULS:TRAN:TRA %s"%(value_f)
        return self.simple_command(ag_81160_id,command)

    def set_symmetry(self,ag_81160_id,symmetry):
        if symmetry=="undef":
            return 1,"ok"
        command="FUNC{channel}:RAMP:SYMM %.9e"%(float(symmetry))
        return self.simple_command(ag_81160_id,command)

    def set_phase(self,ag_81160_id,phase):
        if phase=="undef":
            return 1,"ok"
        command=r"FUNC{channel}:PULSE:DEL:UNIT DEG\n"
        command+="FUNC{channel}:PULSE:DEL %.9e"%(float(phase))
        return self.simple_command(ag_81160_id,command)

    def set_delay(self,ag_81160_id,delay):
        if delay=="undef":
            return 1,"ok"
        command=r"FUNC{channel}:PULSE:DEL:UNIT SEC\n"
        command+="FUNC{channel}:PULSE:DEL %.9e"%(float(delay))
        return self.simple_command(ag_81160_id,command)

    def set_sync(self,ag_81160_id,sync):
        if sync=="undef":
            return 1,"ok"
        command=r"OUTP{channel}:TRIG:ROUT "
        command+=("SYNA" if channel=="1" else "SYNB") if sync=="1" else "NONE"
        return self.simple_command(ag_81160_id,command)

    def power_on(self,ag_81160_id):
        command="OUTP{channel} ON"
        return self.simple_command(ag_81160_id,command)

    def power_off(self,ag_81160_id):
        command="OUTP{channel} OFF"
        return self.simple_command(ag_81160_id,command)

me=ag_81160_class()

def init_ag_81160(ag_81160_id,conf_string):
    """Initialize ag_81160 pattern generator.
    
    *conf_string*: must include:

    - channel: channel on which the ag_81160_id will act
    - bus: conf_string of the underlying link module (gpib, tcp, serial, ...)"""
    return me.init(ag_81160_id,conf_string)

def deinit_ag_81160(ag_81160_id):
    "Deregister an ag_81160 from the pool"
    return me.deinit(ag_81160_id)

def config_ag_81160(ag_81160_id):
    "Configure an ag_81160"
    return me.config(ag_81160_id) 

def inval_ag_81160(ag_81160_id):
    "Invalidate configuration of an ag_81160"
    return me.inval(ag_81160_id)

def reset_ag_81160(ag_81160_id):
    "Send RST signal to PG"
    return me.reset(ag_81160_id)

def configure_base(ag_81160_id,frequency,sync,high_level,low_level):
    retcode,res=me.set_frequency(ag_81160_id,frequency)
    if retcode==0:
        return 0,res
    retcode,res=me.set_sync(ag_81160_id,sync)
    if retcode==0:
        return 0,res
    retcode,res=me.set_high_level(ag_81160_id,high_level)
    if retcode==0:
        return 0,res
    retcode,res=me.set_low_level(ag_81160_id,low_level)
    if retcode==0:
        return 0,res
    return 1,"ok"

def configure_sine_ag_81160(ag_81160_id,frequency,sync,high_level,low_level,phase):
    "Configure ag_81160 to a sine wave with the specified parameters"
    retcode,res=me.set_function(ag_81160_id,"sine")
    if retcode==0:
        return 0,res
    retcode,res=configure_base(ag_81160_id,frequency,sync,high_level,low_level)
    if retcode==0:
        return 0,res
    retcode,res=me.set_phase(ag_81160_id,phase)
    if retcode==0:
        return 0,res
    return 1,"ok"

def configure_square_ag_81160(ag_81160_id,frequency,sync,high_level,low_level,phase,duty_cycle):
    "Configure ag_81160 to a square wave with the specified parameters"
    retcode,res=me.set_function(ag_81160_id,"square")
    if retcode==0:
        return 0,res
    retcode,res=configure_base(ag_81160_id,frequency,sync,high_level,low_level)
    if retcode==0:
        return 0,res
    retcode,res=me.set_phase(ag_81160_id,phase)
    if retcode==0:
        return 0,res
    retcode,res=me.set_duty_cycle(ag_81160_id,duty_cycle)
    if retcode==0:
        return 0,res
    return 1,"ok"

def configure_ramp_ag_81160(ag_81160_id,frequency,sync,high_level,low_level,phase,symmetry):
    "Configure ag_81160 to a ramp wave with the specified parameters"
    retcode,res=me.set_function(ag_81160_id,"ramp")
    if retcode==0:
        return 0,res
    retcode,res=configure_base(ag_81160_id,frequency,sync,high_level,low_level)
    if retcode==0:
        return 0,res
    retcode,res=me.set_phase(ag_81160_id,phase)
    if retcode==0:
        return 0,res
    retcode,res=me.set_symmetry(ag_81160_id,symmetry)
    if retcode==0:
        return 0,res
    return 1,"ok"

def configure_pulse_ag_81160(ag_81160_id,frequency,sync,high_level,low_level,delay,pulse_width,rising_edge,falling_edge):
    "Configure ag_81160 to a pulse train with the specified parameters"
    retcode,res=me.set_function(ag_81160_id,"pulse")
    if retcode==0:
        return 0,res
    retcode,res=configure_base(ag_81160_id,frequency,sync,high_level,low_level)
    if retcode==0:
        return 0,res
    retcode,res=me.set_delay(ag_81160_id,delay)
    if retcode==0:
        return 0,res
    retcode,res=me.set_pulse_width(ag_81160_id,pulse_width)
    if retcode==0:
        return 0,res
    retcode,res=me.set_edges(ag_81160_id,rising_edge,falling_edge)
    if retcode==0:
        return 0,res
    return 1,"ok"

def power_on_ag_81160(ag_81160_id):
    "Turn on"
    return me.power_on(ag_81160_id)

def power_off_ag_81160(ag_81160_id):
    "Turn off"
    return me.power_off(ag_81160_id)
