/*
Copyright 2012-2017 Frédéric Magniette, Miguel Rubio-Roy
This file is part of Pyrame.

Pyrame is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Pyrame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Pyrame.  If not, see <http://www.gnu.org/licenses/>
*/

#include "cmd_gen.h"

#define max(a,b) ((a)>(b)?(a):(b))
#define EVENT_SIZE 40

struct timeval now;

void conv_usage() {
  printf("parameters : dispatcher_nb_blocks\n");
  submod_exit(1);
}

//the thread that generates
void *gen(void *args) {
  struct gen_ws *ws=(struct gen_ws *)args;
  int block_nb=1;
  char block_nb_str[48];
  struct block *b;
  struct event *e;
  int i;
  struct timeval start,end;
  struct sched_param schedp;
  memset(&schedp,0,sizeof(schedp));
  schedp.sched_priority=80;
  gettimeofday(&start,NULL);
  while (block_nb<=ws->total_blocks) {
    //gettimeofday(&now,NULL);
    //printf("generating block %d at %ld.%06ld\n",block_nb,now.tv_sec,now.tv_usec);
    submod_new_block(block_nb,&b);
    // 10 byte block number (fits any 32 bit number)
    sprintf(block_nb_str,"%010d",block_nb);
    submod_set_field_block(b,ws->ese_a,block_nb_str);
    for (i=0;i<ws->nb_events;i++) {
      submod_new_event(b,&e);
      //  !100.000000000|10.000000000|1.0000000000 makes a EVENT_SIZE byte event
      submod_set_field_event(e,ws->ese_b,"100.000000000");
      submod_set_field_event(e,ws->ese_c,"10.000000000");
      submod_set_field_event(e,ws->ese_d,"1.0000000000");
    }
    submod_finalize_block(b);
    block_nb++;
    if (ws->period) {
      // put the thread in soft real-time mode during sleep
      // to be able to do usleeps under 50 us
      sched_setscheduler(0,SCHED_RR,&schedp);
      usleep(ws->period);
      sched_setscheduler(0,SCHED_OTHER,&schedp);
    }
  }
  gettimeofday(&end,NULL);
  ws->last_speed=ws->total_blocks/(end.tv_sec-start.tv_sec+(end.tv_usec-start.tv_usec)/1e6);
  printf("gen thread ending\n");
  return NULL;
}


void *init(int argc,char **argv) {
  int res;
  //init the workspace
  struct gen_ws * ws=malloc(sizeof(struct gen_ws));
  memset(ws,0,sizeof(struct gen_ws));

  //extract parameters from command line
  if (argc<1) {
    printf("1 args mandatory, %d provided\n",argc);
    conv_usage();
  }
  ws->schema=saf_parse_chans("a!b|c|d");
  ws->ese_a=get_ese(ws->schema,"a");
  ws->ese_b=get_ese(ws->schema,"b");
  ws->ese_c=get_ese(ws->schema,"c");
  ws->ese_d=get_ese(ws->schema,"d");
  res=submod_init_dispatcher(atoi(argv[0]),"gen","a!b|c|d","35000");
  if (res<0) {
    printf("error : %s\n",pyrame_errors[-res]);
    submod_exit(1);
  }
  return ws; 
}


void deinit(void *workspace) {
  struct gen_ws *ws=(struct gen_ws *)workspace;
  printf("gen deinit\n");
  if (ws->t_gen!=NULL)
    pthread_cancel(*ws->t_gen);
  printf("gen exiting\n");
}

//pyrapi:flush_gen:
struct cmd_result *flush_gen(struct cmd *command,void *workspace) {
  printf("flush gen\n");
  submod_flush_dispatcher();
  return raw_build_result(1,"ok");
}

//pyrapi:start_gen:block_size,nb_blocks,period
struct cmd_result * start_gen(struct cmd *command,void *workspace) {
  struct gen_ws *ws=(struct gen_ws *)workspace;
  int block_size;

  if (command->nb_params<3) {
    return build_result(0,"gen: 3 arg needeed: %d provided",command->nb_params);
  }

  //block size
  block_size=atoi(command->params[0]);
  if (block_size<EVENT_SIZE) {
    printf("warning: setting block_size to 10\n");
    block_size=10;
  }
  if ((block_size-10)%EVENT_SIZE!=0) {
    block_size-=(block_size-10)%EVENT_SIZE;
    printf("warning: setting block_size to multiple of %d: %d\n",EVENT_SIZE,block_size);
  }

  //blocks to generate
  ws->total_blocks=atoi(command->params[1]);
  if (ws->total_blocks<=0) {
    return raw_build_result(0,"gen: invalid nb of blocks");
  }

  //period
  ws->period=atoi(command->params[2]);

  //nb of events
  ws->nb_events=(block_size-10)/EVENT_SIZE;
  printf("nb_events:%d, for block_size:%d\n",ws->nb_events,block_size);

  //flush internal circular buffer
  submod_flush_dispatcher();

  ws->last_speed=0;
  //launching generator thread
  printf("lanching gen thread\n");
  ws->t_gen=(pthread_t*)malloc(sizeof(pthread_t));
  if (pthread_create(ws->t_gen,NULL,gen,(void *)ws)<0) {
    perror("pthread_create");
    exit(1);
  }
  pthread_detach(*ws->t_gen);
  
  return raw_build_result(1,"ok");
}

//pyrapi:start_sync_gen:block_size,nb_blocks
struct cmd_result * start_sync_gen(struct cmd *command,void *workspace) {
  struct gen_ws *ws=(struct gen_ws *)workspace;
  int block_size;

  if (command->nb_params<3) {
    return build_result(0,"gen: 3 arg needeed: %d provided",command->nb_params);
  }

  //block size
  block_size=atoi(command->params[0]);
  if (block_size<EVENT_SIZE) {
    printf("warning: setting block_size to 10\n");
    block_size=10;
  }
  if ((block_size-10)%EVENT_SIZE!=0) {
    block_size-=(block_size-10)%EVENT_SIZE;
    printf("warning: setting block_size to multiple of %d: %d\n",EVENT_SIZE,block_size);
  }

  //blocks to generate
  ws->total_blocks=atoi(command->params[1]);
  if (ws->total_blocks<=0) {
    return raw_build_result(0,"gen: invalid nb of blocks");
  }

  //period
  ws->period=0;

  //nb of events
  ws->nb_events=(block_size-10)/EVENT_SIZE;
  printf("nb_events:%d, for block_size:%d\n",ws->nb_events,block_size);

  //flush internal circular buffer
  submod_flush_dispatcher();

  ws->last_speed=0;
  gen((void *)ws);
  return raw_build_result(1,"ok");
}

//pyrapi:get_last_speed_gen:
struct cmd_result *get_last_speed_gen(struct cmd *command,void *workspace) {
  struct gen_ws *ws=(struct gen_ws *)workspace;
  return build_result(1,"%f",ws->last_speed);
}

